# Copyright (c) 2019 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import testtools

from kmip.core import enums
from kmip.core import objects
from kmip.core import primitives
from kmip.core import utils

from kmip.core.messages import payloads


class TestLocateRequestPayload(testtools.TestCase):

    def setUp(self):
        super(TestLocateRequestPayload, self).setUp()

        # Encoding obtained from the KMIP 1.1 testing document, Section 15.3.4.
        # Modified to include the Offset Items and Storage Status Mask fields.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Maximum Items - 1
        #     Offset Items - 1
        #     Storage Status Mask - Online Storage | Archival Storage
        #     Object Group Member - Group Member Default
        #     Attribute
        #         Attribute Name - Object Group
        #         Attribute Value - RoundRobinTestGroup
        self.full_encoding = utils.BytearrayStream(
            b'\x42\x00\x79\x01\x00\x00\x00\x80'
            b'\x42\x00\x4F\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\xD4\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x8E\x02\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x00'
            b'\x42\x00\xAC\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x08\x01\x00\x00\x00\x38'
            b'\x42\x00\x0A\x07\x00\x00\x00\x0C'
            b'\x4F\x62\x6A\x65\x63\x74\x20\x47\x72\x6F\x75\x70\x00\x00\x00\x00'
            b'\x42\x00\x0B\x07\x00\x00\x00\x13'
            b'\x52\x6F\x75\x6E\x64\x52\x6F\x62\x69\x6E\x54\x65\x73\x74\x47\x72'
            b'\x6F\x75\x70\x00\x00\x00\x00\x00'
        )

        # Encoding obtained from the KMIP 1.1 testing document, Section 13.3.5.
        # Modified to include the Offset Items, Storage Status Mask, and Object
        # Group Member fields. Manually converted to the KMIP 2.0 format.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Maximum Items - 1
        #     Offset Items - 1
        #     Storage Status Mask - Online Storage | Archival Storage
        #     Object Group Member - Group Member Default
        #     Attributes
        #         Object Type - Public Key
        self.full_encoding_with_attributes = utils.BytearrayStream(
            b'\x42\x00\x79\x01\x00\x00\x00\x58'
            b'\x42\x00\x4F\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\xD4\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x8E\x02\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x00'
            b'\x42\x00\xAC\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x01\x25\x01\x00\x00\x00\x10'
            b'\x42\x00\x57\x05\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x00'
        )

        # Encoding obtained from the KMIP 1.1 testing document, Section 15.3.4.
        # Modified to include the Offset Items and Storage Status Mask fields.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Offset Items - 1
        #     Storage Status Mask - Online Storage | Archival Storage
        #     Object Group Member - Group Member Default
        #     Attribute
        #         Attribute Name - Object Group
        #         Attribute Value - RoundRobinTestGroup
        self.no_maximum_items_encoding = utils.BytearrayStream(
            b'\x42\x00\x79\x01\x00\x00\x00\x70'
            b'\x42\x00\xD4\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x8E\x02\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x00'
            b'\x42\x00\xAC\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x08\x01\x00\x00\x00\x38'
            b'\x42\x00\x0A\x07\x00\x00\x00\x0C'
            b'\x4F\x62\x6A\x65\x63\x74\x20\x47\x72\x6F\x75\x70\x00\x00\x00\x00'
            b'\x42\x00\x0B\x07\x00\x00\x00\x13'
            b'\x52\x6F\x75\x6E\x64\x52\x6F\x62\x69\x6E\x54\x65\x73\x74\x47\x72'
            b'\x6F\x75\x70\x00\x00\x00\x00\x00'
        )

        # Encoding obtained from the KMIP 1.1 testing document, Section 15.3.4.
        # Modified to include the Storage Status Mask field.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Maximum Items - 1
        #     Storage Status Mask - Online Storage | Archival Storage
        #     Object Group Member - Group Member Default
        #     Attribute
        #         Attribute Name - Object Group
        #         Attribute Value - RoundRobinTestGroup
        self.no_offset_items_encoding = utils.BytearrayStream(
            b'\x42\x00\x79\x01\x00\x00\x00\x70'
            b'\x42\x00\x4F\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x8E\x02\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x00'
            b'\x42\x00\xAC\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x08\x01\x00\x00\x00\x38'
            b'\x42\x00\x0A\x07\x00\x00\x00\x0C'
            b'\x4F\x62\x6A\x65\x63\x74\x20\x47\x72\x6F\x75\x70\x00\x00\x00\x00'
            b'\x42\x00\x0B\x07\x00\x00\x00\x13'
            b'\x52\x6F\x75\x6E\x64\x52\x6F\x62\x69\x6E\x54\x65\x73\x74\x47\x72'
            b'\x6F\x75\x70\x00\x00\x00\x00\x00'
        )

        # Encoding obtained from the KMIP 1.1 testing document, Section 15.3.4.
        # Modified to include the Offset Items field.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Maximum Items - 1
        #     Offset Items - 1
        #     Object Group Member - Group Member Default
        #     Attribute
        #         Attribute Name - Object Group
        #         Attribute Value - RoundRobinTestGroup
        self.no_storage_status_mask_encoding = utils.BytearrayStream(
            b'\x42\x00\x79\x01\x00\x00\x00\x70'
            b'\x42\x00\x4F\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\xD4\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\xAC\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
            b'\x42\x00\x08\x01\x00\x00\x00\x38'
            b'\x42\x00\x0A\x07\x00\x00\x00\x0C'
            b'\x4F\x62\x6A\x65\x63\x74\x20\x47\x72\x6F\x75\x70\x00\x00\x00\x00'
            b'\x42\x00\x0B\x07\x00\x00\x00\x13'
            b'\x52\x6F\x75\x6E\x64\x52\x6F\x62\x69\x6E\x54\x65\x73\x74\x47\x72'
            b'\x6F\x75\x70\x00\x00\x00\x00\x00'
        )

        # Encoding obtained from the KMIP 1.1 testing document, Section 15.3.4.
        # Modified to include the Offset Items and Storage Status Mask fields.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Maximum Items - 1
        #     Offset Items - 1
        #     Storage Status Mask - Online Storage | Archival Storage
        #     Attribute
        #         Attribute Name - Object Group
        #         Attribute Value - RoundRobinTestGroup
        self.no_object_group_member_encoding = utils.BytearrayStream(
            b'\x42\x00\x79\x01\x00\x00\x00\x70'
            b'\x42\x00\x4F\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\xD4\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x8E\x02\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x00'
            b'\x42\x00\x08\x01\x00\x00\x00\x38'
            b'\x42\x00\x0A\x07\x00\x00\x00\x0C'
            b'\x4F\x62\x6A\x65\x63\x74\x20\x47\x72\x6F\x75\x70\x00\x00\x00\x00'
            b'\x42\x00\x0B\x07\x00\x00\x00\x13'
            b'\x52\x6F\x75\x6E\x64\x52\x6F\x62\x69\x6E\x54\x65\x73\x74\x47\x72'
            b'\x6F\x75\x70\x00\x00\x00\x00\x00'
        )

        # Encoding obtained from the KMIP 1.1 testing document, Section 15.3.4.
        # Modified to include the Offset Items and Storage Status Mask fields.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Maximum Items - 1
        #     Offset Items - 1
        #     Storage Status Mask - Online Storage | Archival Storage
        #     Object Group Member - Group Member Default
        self.no_attributes_encoding = utils.BytearrayStream(
            b'\x42\x00\x79\x01\x00\x00\x00\x40'
            b'\x42\x00\x4F\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\xD4\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x8E\x02\x00\x00\x00\x04\x00\x00\x00\x03\x00\x00\x00\x00'
            b'\x42\x00\xAC\x05\x00\x00\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00'
        )

        # This encoding matches the following set of values:
        # Request Payload
        self.empty_encoding = utils.BytearrayStream(
            b'\x42\x00\x79\x01\x00\x00\x00\x00'
        )

    def tearDown(self):
        super(TestLocateRequestPayload, self).tearDown()

    def test_invalid_maximum_items(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the maximum items of a Locate request payload.
        """
        kwargs = {"maximum_items": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "Maximum items must be an integer.",
            payloads.LocateRequestPayload,
            **kwargs
        )

        args = (
            payloads.LocateRequestPayload(),
            "maximum_items",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "Maximum items must be an integer.",
            setattr,
            *args
        )

    def test_invalid_offset_items(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the offset items of a Locate request payload.
        """
        kwargs = {"offset_items": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "Offset items must be an integer.",
            payloads.LocateRequestPayload,
            **kwargs
        )

        args = (
            payloads.LocateRequestPayload(),
            "offset_items",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "Offset items must be an integer.",
            setattr,
            *args
        )

    def test_invalid_storage_status_mask(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the storage status mask of a Locate request payload.
        """
        kwargs = {"storage_status_mask": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "Storage status mask must be an integer representing a valid "
            "StorageStatusMask bit mask.",
            payloads.LocateRequestPayload,
            **kwargs
        )

        kwargs = {"storage_status_mask": 55}
        self.assertRaisesRegex(
            TypeError,
            "Storage status mask must be an integer representing a valid "
            "StorageStatusMask bit mask.",
            payloads.LocateRequestPayload,
            **kwargs
        )

        args = (
            payloads.LocateRequestPayload(),
            "storage_status_mask",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "Storage status mask must be an integer representing a valid "
            "StorageStatusMask bit mask.",
            setattr,
            *args
        )

    def test_invalid_object_group_member(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the object group member of a Locate request payload.
        """
        kwargs = {"object_group_member": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "Object group member must be an ObjectGroupMember enumeration.",
            payloads.LocateRequestPayload,
            **kwargs
        )

        args = (
            payloads.LocateRequestPayload(),
            "object_group_member",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "Object group member must be an ObjectGroupMember enumeration.",
            setattr,
            *args
        )

    def test_invalid_attributes(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the attributes of a Locate request payload.
        """
        kwargs = {"attributes": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "Attributes must be a list of Attribute structures.",
            payloads.LocateRequestPayload,
            **kwargs
        )

        kwargs = {"attributes": ["invalid"]}
        self.assertRaisesRegex(
            TypeError,
            "Attributes must be a list of Attribute structures.",
            payloads.LocateRequestPayload,
            **kwargs
        )

        args = (
            payloads.LocateRequestPayload(),
            "attributes",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "Attributes must be a list of Attribute structures.",
            setattr,
            *args
        )

        args = (
            payloads.LocateRequestPayload(),
            "attributes",
            ["invalid"]
        )
        self.assertRaisesRegex(
            TypeError,
            "Attributes must be a list of Attribute structures.",
            setattr,
            *args
        )

    def test_read(self):
        """
        Test that a Locate request payload can be read from a data stream.
        """
        payload = payloads.LocateRequestPayload()

        self.assertIsNone(payload.maximum_items)
        self.assertIsNone(payload.offset_items)
        self.assertIsNone(payload.storage_status_mask)
        self.assertIsNone(payload.object_group_member)
        self.assertEqual([], payload.attributes)

        payload.read(self.full_encoding)

        self.assertEqual(1, payload.maximum_items)
        self.assertEqual(1, payload.offset_items)
        self.assertEqual(
            enums.get_bit_mask_from_enumerations(
                [
                    enums.StorageStatusMask.ONLINE_STORAGE,
                    enums.StorageStatusMask.ARCHIVAL_STORAGE
                ]
            ),
            payload.storage_status_mask
        )
        self.assertEqual(
            enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            payload.object_group_member
        )
        self.assertIsInstance(payload.attributes, list)
        self.assertEqual(1, len(payload.attributes))
        self.assertEqual(
            objects.Attribute(
                attribute_name=objects.Attribute.AttributeName("Object Group"),
                attribute_value=primitives.TextString(
                    value="RoundRobinTestGroup",
                    tag=enums.Tags.OBJECT_GROUP
                )
            ),
            payload.attributes[0]
        )

    def test_read_kmip_2_0(self):
        """
        Test that a Locate request payload can be read from a data stream
        encoded with the KMIP 2.0 format.
        """
        payload = payloads.LocateRequestPayload()

        self.assertIsNone(payload.maximum_items)
        self.assertIsNone(payload.offset_items)
        self.assertIsNone(payload.storage_status_mask)
        self.assertIsNone(payload.object_group_member)
        self.assertEqual([], payload.attributes)

        payload.read(
            self.full_encoding_with_attributes,
            kmip_version=enums.KMIPVersion.KMIP_2_0
        )

        self.assertEqual(1, payload.maximum_items)
        self.assertEqual(1, payload.offset_items)
        self.assertEqual(
            enums.get_bit_mask_from_enumerations(
                [
                    enums.StorageStatusMask.ONLINE_STORAGE,
                    enums.StorageStatusMask.ARCHIVAL_STORAGE
                ]
            ),
            payload.storage_status_mask
        )
        self.assertEqual(
            enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            payload.object_group_member
        )
        self.assertIsInstance(payload.attributes, list)
        self.assertEqual(1, len(payload.attributes))
        self.assertEqual(
            objects.Attribute(
                attribute_name=objects.Attribute.AttributeName("Object Type"),
                attribute_value=primitives.Enumeration(
                    enums.ObjectType,
                    value=enums.ObjectType.PUBLIC_KEY,
                    tag=enums.Tags.OBJECT_TYPE
                )
            ),
            payload.attributes[0]
        )

    def test_read_missing_maximum_items(self):
        """
        Test that a Locate request payload can be read from a data stream
        even when missing the maximum items.
        """
        payload = payloads.LocateRequestPayload()

        self.assertIsNone(payload.maximum_items)
        self.assertIsNone(payload.offset_items)
        self.assertIsNone(payload.storage_status_mask)
        self.assertIsNone(payload.object_group_member)
        self.assertEqual([], payload.attributes)

        payload.read(self.no_maximum_items_encoding)

        self.assertIsNone(payload.maximum_items)
        self.assertEqual(1, payload.offset_items)
        self.assertEqual(
            enums.get_bit_mask_from_enumerations(
                [
                    enums.StorageStatusMask.ONLINE_STORAGE,
                    enums.StorageStatusMask.ARCHIVAL_STORAGE
                ]
            ),
            payload.storage_status_mask
        )
        self.assertEqual(
            enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            payload.object_group_member
        )
        self.assertIsInstance(payload.attributes, list)
        self.assertEqual(1, len(payload.attributes))
        self.assertEqual(
            objects.Attribute(
                attribute_name=objects.Attribute.AttributeName("Object Group"),
                attribute_value=primitives.TextString(
                    value="RoundRobinTestGroup",
                    tag=enums.Tags.OBJECT_GROUP
                )
            ),
            payload.attributes[0]
        )

    def test_read_missing_offset_items(self):
        """
        Test that a Locate request payload can be read from a data stream
        even when missing the offset items.
        """
        payload = payloads.LocateRequestPayload()

        self.assertIsNone(payload.maximum_items)
        self.assertIsNone(payload.offset_items)
        self.assertIsNone(payload.storage_status_mask)
        self.assertIsNone(payload.object_group_member)
        self.assertEqual([], payload.attributes)

        payload.read(self.no_offset_items_encoding)

        self.assertEqual(1, payload.maximum_items)
        self.assertIsNone(payload.offset_items)
        self.assertEqual(
            enums.get_bit_mask_from_enumerations(
                [
                    enums.StorageStatusMask.ONLINE_STORAGE,
                    enums.StorageStatusMask.ARCHIVAL_STORAGE
                ]
            ),
            payload.storage_status_mask
        )
        self.assertEqual(
            enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            payload.object_group_member
        )
        self.assertIsInstance(payload.attributes, list)
        self.assertEqual(1, len(payload.attributes))
        self.assertEqual(
            objects.Attribute(
                attribute_name=objects.Attribute.AttributeName("Object Group"),
                attribute_value=primitives.TextString(
                    value="RoundRobinTestGroup",
                    tag=enums.Tags.OBJECT_GROUP
                )
            ),
            payload.attributes[0]
        )

    def test_read_missing_storage_status_mask(self):
        """
        Test that a Locate request payload can be read from a data stream
        even when missing the storage status mask.
        """
        payload = payloads.LocateRequestPayload()

        self.assertIsNone(payload.maximum_items)
        self.assertIsNone(payload.offset_items)
        self.assertIsNone(payload.storage_status_mask)
        self.assertIsNone(payload.object_group_member)
        self.assertEqual([], payload.attributes)

        payload.read(self.no_storage_status_mask_encoding)

        self.assertEqual(1, payload.maximum_items)
        self.assertEqual(1, payload.offset_items)
        self.assertIsNone(payload.storage_status_mask)
        self.assertEqual(
            enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            payload.object_group_member
        )
        self.assertIsInstance(payload.attributes, list)
        self.assertEqual(1, len(payload.attributes))
        self.assertEqual(
            objects.Attribute(
                attribute_name=objects.Attribute.AttributeName("Object Group"),
                attribute_value=primitives.TextString(
                    value="RoundRobinTestGroup",
                    tag=enums.Tags.OBJECT_GROUP
                )
            ),
            payload.attributes[0]
        )

    def test_read_missing_object_group_member(self):
        """
        Test that a Locate request payload can be read from a data stream
        even when missing the object group member.
        """
        payload = payloads.LocateRequestPayload()

        self.assertIsNone(payload.maximum_items)
        self.assertIsNone(payload.offset_items)
        self.assertIsNone(payload.storage_status_mask)
        self.assertIsNone(payload.object_group_member)
        self.assertEqual([], payload.attributes)

        payload.read(self.no_object_group_member_encoding)

        self.assertEqual(1, payload.maximum_items)
        self.assertEqual(1, payload.offset_items)
        self.assertEqual(
            enums.get_bit_mask_from_enumerations(
                [
                    enums.StorageStatusMask.ONLINE_STORAGE,
                    enums.StorageStatusMask.ARCHIVAL_STORAGE
                ]
            ),
            payload.storage_status_mask
        )
        self.assertIsNone(payload.object_group_member)
        self.assertIsInstance(payload.attributes, list)
        self.assertEqual(1, len(payload.attributes))
        self.assertEqual(
            objects.Attribute(
                attribute_name=objects.Attribute.AttributeName("Object Group"),
                attribute_value=primitives.TextString(
                    value="RoundRobinTestGroup",
                    tag=enums.Tags.OBJECT_GROUP
                )
            ),
            payload.attributes[0]
        )

    def test_read_missing_attributes(self):
        """
        Test that a Locate request payload can be read from a data stream
        even when missing the attributes.
        """
        payload = payloads.LocateRequestPayload()

        self.assertIsNone(payload.maximum_items)
        self.assertIsNone(payload.offset_items)
        self.assertIsNone(payload.storage_status_mask)
        self.assertIsNone(payload.object_group_member)
        self.assertEqual([], payload.attributes)

        payload.read(self.no_attributes_encoding)

        self.assertEqual(1, payload.maximum_items)
        self.assertEqual(1, payload.offset_items)
        self.assertEqual(
            enums.get_bit_mask_from_enumerations(
                [
                    enums.StorageStatusMask.ONLINE_STORAGE,
                    enums.StorageStatusMask.ARCHIVAL_STORAGE
                ]
            ),
            payload.storage_status_mask
        )
        self.assertEqual(
            enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            payload.object_group_member
        )
        self.assertEqual([], payload.attributes)

    def test_read_missing_everything(self):
        """
        Test that a Locate request payload can be read from a data stream
        even when missing all fields.
        """
        payload = payloads.LocateRequestPayload()

        self.assertIsNone(payload.maximum_items)
        self.assertIsNone(payload.offset_items)
        self.assertIsNone(payload.storage_status_mask)
        self.assertIsNone(payload.object_group_member)
        self.assertEqual([], payload.attributes)

        payload.read(self.empty_encoding)

        self.assertIsNone(payload.maximum_items)
        self.assertIsNone(payload.offset_items)
        self.assertIsNone(payload.storage_status_mask)
        self.assertIsNone(payload.object_group_member)
        self.assertEqual([], payload.attributes)

    def test_write(self):
        """
        Test that a Locate request payload can be written to a data stream.
        """
        payload = payloads.LocateRequestPayload(
            maximum_items=1,
            offset_items=1,
            storage_status_mask=3,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )

        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(len(self.full_encoding), len(stream))
        self.assertEqual(str(self.full_encoding), str(stream))

    def test_write_kmip_2_0(self):
        """
        Test that a Locate request payload can be written to a data stream
        encoded with the KMIP 2.0 format.
        """
        payload = payloads.LocateRequestPayload(
            maximum_items=1,
            offset_items=1,
            storage_status_mask=3,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Type"
                    ),
                    attribute_value=primitives.Enumeration(
                        enums.ObjectType,
                        value=enums.ObjectType.PUBLIC_KEY,
                        tag=enums.Tags.OBJECT_TYPE
                    )
                )
            ]
        )

        stream = utils.BytearrayStream()
        payload.write(stream, kmip_version=enums.KMIPVersion.KMIP_2_0)

        self.assertEqual(len(self.full_encoding_with_attributes), len(stream))
        self.assertEqual(str(self.full_encoding_with_attributes), str(stream))

    def test_write_missing_maximum_items(self):
        """
        Test that a Locate request payload can be written to a data stream
        even when missing the maximum items.
        """
        payload = payloads.LocateRequestPayload(
            offset_items=1,
            storage_status_mask=3,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )

        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(len(self.no_maximum_items_encoding), len(stream))
        self.assertEqual(str(self.no_maximum_items_encoding), str(stream))

    def test_write_missing_offset_items(self):
        """
        Test that a Locate request payload can be written to a data stream
        even when missing the offset items.
        """
        payload = payloads.LocateRequestPayload(
            maximum_items=1,
            storage_status_mask=3,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )

        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(len(self.no_offset_items_encoding), len(stream))
        self.assertEqual(str(self.no_offset_items_encoding), str(stream))

    def test_write_missing_storage_status_mask(self):
        """
        Test that a Locate request payload can be written to a data stream
        even when missing the storage status mask.
        """
        payload = payloads.LocateRequestPayload(
            maximum_items=1,
            offset_items=1,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )

        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(
            len(self.no_storage_status_mask_encoding),
            len(stream)
        )
        self.assertEqual(
            str(self.no_storage_status_mask_encoding),
            str(stream)
        )

    def test_write_missing_object_group_member(self):
        """
        Test that a Locate request payload can be written to a data stream
        even when missing the object group member.
        """
        payload = payloads.LocateRequestPayload(
            maximum_items=1,
            offset_items=1,
            storage_status_mask=3,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )

        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(
            len(self.no_object_group_member_encoding),
            len(stream)
        )
        self.assertEqual(
            str(self.no_object_group_member_encoding),
            str(stream)
        )

    def test_write_missing_attributes(self):
        """
        Test that a Locate request payload can be written to a data stream
        even when missing the attributes.
        """
        payload = payloads.LocateRequestPayload(
            maximum_items=1,
            offset_items=1,
            storage_status_mask=3,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
        )

        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(len(self.no_attributes_encoding), len(stream))
        self.assertEqual(str(self.no_attributes_encoding), str(stream))

    def test_write_missing_everything(self):
        """
        Test that a Locate request payload can be written to a data stream
        even when missing all fields.
        """
        payload = payloads.LocateRequestPayload()

        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(len(self.empty_encoding), len(stream))
        self.assertEqual(str(self.empty_encoding), str(stream))

    def test_repr(self):
        """
        Test that repr can be applied to a Locate request payload structure.
        """
        payload = payloads.LocateRequestPayload(
            maximum_items=1,
            offset_items=1,
            storage_status_mask=3,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )

        s = "LocateRequestPayload(" \
            "maximum_items=1, " \
            "offset_items=1, " \
            "storage_status_mask=3, " \
            "object_group_member=ObjectGroupMember.GROUP_MEMBER_DEFAULT, " \
            "attributes=["
#            "Attribute(" \
#            "attribute_name=AttributeName(value='Object Group'), " \
#            "attribute_index=None, " \
#            "attribute_value=TextString(value='RoundRobinTestGroup'))" \
#            "])"

        # TODO (ph) Uncomment above when Attribute repr fixed. Fix below too.

        self.assertTrue(repr(payload).startswith(s))

    def str(self):
        """
        Test that str can be applied to a Locate request payload structure.
        """
        payload = payloads.LocateRequestPayload(
            maximum_items=1,
            offset_items=1,
            storage_status_mask=3,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )

        s = '{' \
            '"maximum_items": 1, ' \
            '"offset_items": 1, ' \
            '"storage_status_mask": 3, ' \
            '"object_group_member": ObjectGroupMember.GROUP_MEMBER_DEFAULT, ' \
            '"attributes=[' \
            '{' \
            '"attribute_name": "Object Group", ' \
            '"attribute_index": None, ' \
            '"attribute_value": "RoundRobinTestGroup"' \
            '}]' \
            '}'

        self.assertEqual(s, str(payload))

    def test_equal_on_equal(self):
        """
        Test that the equality operator returns True when comparing two Locate
        request payloads with the same data.
        """
        a = payloads.LocateRequestPayload()
        b = payloads.LocateRequestPayload()

        self.assertTrue(a == b)
        self.assertTrue(b == a)

        a = payloads.LocateRequestPayload(
            maximum_items=1,
            offset_items=1,
            storage_status_mask=3,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )
        b = payloads.LocateRequestPayload(
            maximum_items=1,
            offset_items=1,
            storage_status_mask=3,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )

        self.assertTrue(a == b)
        self.assertTrue(b == a)

    def test_equal_on_not_equal_maximum_items(self):
        """
        Test that the equality operator returns False when comparing two
        Locate request payloads with different maximum items.
        """
        a = payloads.LocateRequestPayload(maximum_items=1)
        b = payloads.LocateRequestPayload(maximum_items=2)

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_not_equal_offset_items(self):
        """
        Test that the equality operator returns False when comparing two
        Locate request payloads with different offset items.
        """
        a = payloads.LocateRequestPayload(offset_items=1)
        b = payloads.LocateRequestPayload(offset_items=2)

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_not_equal_storage_status_mask(self):
        """
        Test that the equality operator returns False when comparing two
        Locate request payloads with different storage status mask.
        """
        a = payloads.LocateRequestPayload(storage_status_mask=1)
        b = payloads.LocateRequestPayload(storage_status_mask=2)

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_not_equal_object_group_member(self):
        """
        Test that the equality operator returns False when comparing two
        Locate request payloads with different object group member.
        """
        a = payloads.LocateRequestPayload(
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT
        )
        b = payloads.LocateRequestPayload(
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_FRESH
        )

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_not_equal_attributes(self):
        """
        Test that the equality operator returns False when comparing two
        Locate request payloads with different attributes.
        """
        a = payloads.LocateRequestPayload(
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )
        b = payloads.LocateRequestPayload(
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Cryptographic Algorithm"
                    ),
                    attribute_value=primitives.Enumeration(
                        enums.CryptographicAlgorithm,
                        value=enums.CryptographicAlgorithm.AES,
                        tag=enums.Tags.CRYPTOGRAPHIC_ALGORITHM
                    )
                )
            ]
        )

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_type_mismatch(self):
        """
        Test that the equality operator returns False when comparing two
        Locate request payloads with different types.
        """
        a = payloads.LocateRequestPayload()
        b = "invalid"

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_not_equal_on_equal(self):
        """
        Test that the inequality operator returns False when comparing two
        Locate request payloads with the same data.
        """
        a = payloads.LocateRequestPayload()
        b = payloads.LocateRequestPayload()

        self.assertFalse(a != b)
        self.assertFalse(b != a)

        a = payloads.LocateRequestPayload(
            maximum_items=1,
            offset_items=1,
            storage_status_mask=3,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )
        b = payloads.LocateRequestPayload(
            maximum_items=1,
            offset_items=1,
            storage_status_mask=3,
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT,
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )

        self.assertFalse(a != b)
        self.assertFalse(b != a)

    def test_not_equal_on_not_equal_maximum_items(self):
        """
        Test that the inequality operator returns True when comparing two
        Locate request payloads with different maximum items.
        """
        a = payloads.LocateRequestPayload(maximum_items=1)
        b = payloads.LocateRequestPayload(maximum_items=2)

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_not_equal_offset_items(self):
        """
        Test that the inequality operator returns True when comparing two
        Locate request payloads with different offset items.
        """
        a = payloads.LocateRequestPayload(offset_items=1)
        b = payloads.LocateRequestPayload(offset_items=2)

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_not_equal_storage_status_mask(self):
        """
        Test that the inequality operator returns True when comparing two
        Locate request payloads with different storage status mask.
        """
        a = payloads.LocateRequestPayload(storage_status_mask=1)
        b = payloads.LocateRequestPayload(storage_status_mask=2)

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_not_equal_object_group_member(self):
        """
        Test that the inequality operator returns True when comparing two
        Locate request payloads with different object group member.
        """
        a = payloads.LocateRequestPayload(
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_DEFAULT
        )
        b = payloads.LocateRequestPayload(
            object_group_member=enums.ObjectGroupMember.GROUP_MEMBER_FRESH
        )

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_not_equal_attributes(self):
        """
        Test that the inequality operator returns True when comparing two
        Locate request payloads with different attributes.
        """
        a = payloads.LocateRequestPayload(
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Object Group"
                    ),
                    attribute_value=primitives.TextString(
                        value="RoundRobinTestGroup",
                        tag=enums.Tags.OBJECT_GROUP
                    )
                )
            ]
        )
        b = payloads.LocateRequestPayload(
            attributes=[
                objects.Attribute(
                    attribute_name=objects.Attribute.AttributeName(
                        "Cryptographic Algorithm"
                    ),
                    attribute_value=primitives.Enumeration(
                        enums.CryptographicAlgorithm,
                        value=enums.CryptographicAlgorithm.AES,
                        tag=enums.Tags.CRYPTOGRAPHIC_ALGORITHM
                    )
                )
            ]
        )

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_type_mismatch(self):
        """
        Test that the inequality operator returns True when comparing two
        Locate request payloads with different types.
        """
        a = payloads.LocateRequestPayload()
        b = "invalid"

        self.assertTrue(a != b)
        self.assertTrue(b != a)


class TestLocateResponsePayload(testtools.TestCase):

    def setUp(self):
        super(TestLocateResponsePayload, self).setUp()

        # Encoding obtained from the KMIP 1.1 testing document, Section 15.3.4.
        # Modified to include the Located Items field.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Located Items - 1
        #     Unique Identifier - 8d945322-fd70-495d-bf7f-71481d1401f6
        self.full_encoding = utils.BytearrayStream(
            b'\x42\x00\x7C\x01\x00\x00\x00\x40'
            b'\x42\x00\xD5\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
            b'\x42\x00\x94\x07\x00\x00\x00\x24'
            b'\x38\x64\x39\x34\x35\x33\x32\x32\x2D\x66\x64\x37\x30\x2D\x34\x39'
            b'\x35\x64\x2D\x62\x66\x37\x66\x2D\x37\x31\x34\x38\x31\x64\x31\x34'
            b'\x30\x31\x66\x36\x00\x00\x00\x00'
        )

        # Encoding obtained from the KMIP 1.1 testing document, Section 15.3.4.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Unique Identifier - 8d945322-fd70-495d-bf7f-71481d1401f6
        self.no_located_items_encoding = utils.BytearrayStream(
            b'\x42\x00\x7C\x01\x00\x00\x00\x30'
            b'\x42\x00\x94\x07\x00\x00\x00\x24'
            b'\x38\x64\x39\x34\x35\x33\x32\x32\x2D\x66\x64\x37\x30\x2D\x34\x39'
            b'\x35\x64\x2D\x62\x66\x37\x66\x2D\x37\x31\x34\x38\x31\x64\x31\x34'
            b'\x30\x31\x66\x36\x00\x00\x00\x00'
        )

        # Encoding obtained from the KMIP 1.1 testing document, Section 15.3.4.
        # Modified to include the Located Items field.
        #
        # This encoding matches the following set of values:
        # Request Payload
        #     Located Items - 1
        self.no_unique_identifiers_encoding = utils.BytearrayStream(
            b'\x42\x00\x7C\x01\x00\x00\x00\x10'
            b'\x42\x00\xD5\x02\x00\x00\x00\x04\x00\x00\x00\x01\x00\x00\x00\x00'
        )

        # This encoding matches the following set of values:
        # Request Payload
        self.empty_encoding = utils.BytearrayStream(
            b'\x42\x00\x7C\x01\x00\x00\x00\x00'
        )

    def tearDown(self):
        super(TestLocateResponsePayload, self).tearDown()

    def test_invalid_located_items(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the located items of a Locate response payload.
        """
        kwargs = {"located_items": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "Located items must be an integer.",
            payloads.LocateResponsePayload,
            **kwargs
        )

        args = (
            payloads.LocateResponsePayload(),
            "located_items",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "Located items must be an integer.",
            setattr,
            *args
        )

    def test_invalid_unique_identifiers(self):
        """
        Test that a TypeError is raised when an invalid value is used to set
        the unique identifiers of a Locate response payload.
        """
        kwargs = {"unique_identifiers": "invalid"}
        self.assertRaisesRegex(
            TypeError,
            "Unique identifiers must be a list of strings.",
            payloads.LocateResponsePayload,
            **kwargs
        )

        kwargs = {"unique_identifiers": [0]}
        self.assertRaisesRegex(
            TypeError,
            "Unique identifiers must be a list of strings.",
            payloads.LocateResponsePayload,
            **kwargs
        )

        args = (
            payloads.LocateResponsePayload(),
            "unique_identifiers",
            "invalid"
        )
        self.assertRaisesRegex(
            TypeError,
            "Unique identifiers must be a list of strings.",
            setattr,
            *args
        )

        args = (
            payloads.LocateResponsePayload(),
            "unique_identifiers",
            [0]
        )
        self.assertRaisesRegex(
            TypeError,
            "Unique identifiers must be a list of strings.",
            setattr,
            *args
        )

    def test_read(self):
        """
        Test that a Locate response payload can be read from a data stream.
        """
        payload = payloads.LocateResponsePayload()

        self.assertIsNone(payload.located_items)
        self.assertEqual([], payload.unique_identifiers)

        payload.read(self.full_encoding)

        self.assertEqual(1, payload.located_items)
        self.assertIsInstance(payload.unique_identifiers, list)
        self.assertEqual(1, len(payload.unique_identifiers))
        self.assertEqual(
            ["8d945322-fd70-495d-bf7f-71481d1401f6"],
            payload.unique_identifiers
        )

    def test_read_missing_located_items(self):
        """
        Test that a Locate response payload can be read from a data stream
        even when missing the located items.
        """
        payload = payloads.LocateResponsePayload()

        self.assertIsNone(payload.located_items)
        self.assertEqual([], payload.unique_identifiers)

        payload.read(self.no_located_items_encoding)

        self.assertIsNone(payload.located_items)
        self.assertIsInstance(payload.unique_identifiers, list)
        self.assertEqual(1, len(payload.unique_identifiers))
        self.assertEqual(
            ["8d945322-fd70-495d-bf7f-71481d1401f6"],
            payload.unique_identifiers
        )

    def test_read_missing_unique_identifiers(self):
        """
        Test that a Locate response payload can be read from a data stream
        even when missing the unique identifiers.
        """
        payload = payloads.LocateResponsePayload()

        self.assertIsNone(payload.located_items)
        self.assertEqual([], payload.unique_identifiers)

        payload.read(self.no_unique_identifiers_encoding)

        self.assertEqual(1, payload.located_items)
        self.assertEqual([], payload.unique_identifiers)

    def test_read_missing_everything(self):
        """
        Test that a Locate response payload can be read from a data stream
        even when missing all fields.
        """
        payload = payloads.LocateResponsePayload()

        self.assertIsNone(payload.located_items)
        self.assertEqual([], payload.unique_identifiers)

        payload.read(self.empty_encoding)

        self.assertIsNone(payload.located_items)
        self.assertEqual([], payload.unique_identifiers)

    def test_write(self):
        """
        Test that a Locate response payload can be written to a data stream.
        """
        payload = payloads.LocateResponsePayload(
            located_items=1,
            unique_identifiers=["8d945322-fd70-495d-bf7f-71481d1401f6"]
        )

        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(len(self.full_encoding), len(stream))
        self.assertEqual(str(self.full_encoding), str(stream))

    def test_write_missing_located_items(self):
        """
        Test that a Locate response payload can be written to a data stream
        even when missing the located items.
        """
        payload = payloads.LocateResponsePayload(
            located_items=1
        )

        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(len(self.no_unique_identifiers_encoding), len(stream))
        self.assertEqual(str(self.no_unique_identifiers_encoding), str(stream))

    def test_write_missing_unique_identifiers(self):
        """
        Test that a Locate response payload can be written to a data stream
        even when missing the unique identifiers.
        """
        payload = payloads.LocateResponsePayload(
            located_items=1,
            unique_identifiers=["8d945322-fd70-495d-bf7f-71481d1401f6"]
        )

        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(len(self.full_encoding), len(stream))
        self.assertEqual(str(self.full_encoding), str(stream))

    def test_write_missing_everything(self):
        """
        Test that a Locate response payload can be written to a data stream
        even when missing all fields.
        """
        payload = payloads.LocateResponsePayload()

        stream = utils.BytearrayStream()
        payload.write(stream)

        self.assertEqual(len(self.empty_encoding), len(stream))
        self.assertEqual(str(self.empty_encoding), str(stream))

    def test_repr(self):
        """
        Test that repr can be applied to a Locate response payload structure.
        """
        payload = payloads.LocateResponsePayload(
            located_items=1,
            unique_identifiers=["8d945322-fd70-495d-bf7f-71481d1401f6"]
        )

        s = "LocateResponsePayload(" \
            "located_items=1, " \
            "unique_identifiers=['8d945322-fd70-495d-bf7f-71481d1401f6'])"

        self.assertEqual(s, repr(payload))

    def str(self):
        """
        Test that str can be applied to a Locate response payload structure.
        """
        payload = payloads.LocateResponsePayload(
            located_items=1,
            unique_identifiers=["8d945322-fd70-495d-bf7f-71481d1401f6"]
        )

        s = "LocateResponsePayload(" \
            "located_items=1, " \
            "unique_identifiers=['8d945322-fd70-495d-bf7f-71481d1401f6'])"

        self.assertEqual(s, repr(payload))

        s = '{' \
            '"located_items": 1, ' \
            '"unique_identifiers": [' \
            '"8d945322-fd70-495d-bf7f-71481d1401f6"]' \
            ']' \
            '}'

        self.assertEqual(s, str(payload))

    def test_equal_on_equal(self):
        """
        Test that the equality operator returns True when comparing two Locate
        response payloads with the same data.
        """
        a = payloads.LocateResponsePayload()
        b = payloads.LocateResponsePayload()

        self.assertTrue(a == b)
        self.assertTrue(b == a)

        a = payloads.LocateResponsePayload(
            located_items=1,
            unique_identifiers=["8d945322-fd70-495d-bf7f-71481d1401f6"]
        )
        b = payloads.LocateResponsePayload(
            located_items=1,
            unique_identifiers=["8d945322-fd70-495d-bf7f-71481d1401f6"]
        )

        self.assertTrue(a == b)
        self.assertTrue(b == a)

    def test_equal_on_not_equal_located_items(self):
        """
        Test that the equality operator returns False when comparing two
        Locate response payloads with different located items.
        """
        a = payloads.LocateResponsePayload(located_items=1)
        b = payloads.LocateResponsePayload(located_items=2)

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_not_equal_unique_identifiers(self):
        """
        Test that the equality operator returns False when comparing two
        Locate response payloads with different unique identifiers.
        """
        a = payloads.LocateResponsePayload(
            unique_identifiers=["8d945322-fd70-495d-bf7f-71481d1401f6"]
        )
        b = payloads.LocateResponsePayload(
            unique_identifiers=["49a1ca88-6bea-4fb2-b450-7e58802c3038"]
        )

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_equal_on_type_mismatch(self):
        """
        Test that the equality operator returns False when comparing two
        Locate response payloads with different types.
        """
        a = payloads.LocateResponsePayload()
        b = "invalid"

        self.assertFalse(a == b)
        self.assertFalse(b == a)

    def test_not_equal_on_equal(self):
        """
        Test that the inequality operator returns False when comparing two
        Locate response payloads with the same data.
        """
        a = payloads.LocateResponsePayload()
        b = payloads.LocateResponsePayload()

        self.assertFalse(a != b)
        self.assertFalse(b != a)

        a = payloads.LocateResponsePayload(
            located_items=1,
            unique_identifiers=["8d945322-fd70-495d-bf7f-71481d1401f6"]
        )
        b = payloads.LocateResponsePayload(
            located_items=1,
            unique_identifiers=["8d945322-fd70-495d-bf7f-71481d1401f6"]
        )

        self.assertFalse(a != b)
        self.assertFalse(b != a)

    def test_not_equal_on_not_equal_located_items(self):
        """
        Test that the inequality operator returns True when comparing two
        Locate response payloads with different located items.
        """
        a = payloads.LocateResponsePayload(located_items=1)
        b = payloads.LocateResponsePayload(located_items=2)

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_not_equal_unique_identifiers(self):
        """
        Test that the inequality operator returns True when comparing two
        Locate response payloads with different unique identifiers.
        """
        a = payloads.LocateResponsePayload(
            unique_identifiers=["8d945322-fd70-495d-bf7f-71481d1401f6"]
        )
        b = payloads.LocateResponsePayload(
            unique_identifiers=["49a1ca88-6bea-4fb2-b450-7e58802c3038"]
        )

        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_not_equal_on_type_mismatch(self):
        """
        Test that the inequality operator returns True when comparing two
        Locate response payloads with different types.
        """
        a = payloads.LocateResponsePayload()
        b = "invalid"

        self.assertTrue(a != b)
        self.assertTrue(b != a)



# Copyright (c) 2019 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sqlalchemy
import testtools

from kmip.pie import objects
from kmip.pie import sqltypes


class TestObjectGroup(testtools.TestCase):
    """
    Test suite for ObjectGroup.
    """

    def setUp(self):
        super(TestObjectGroup, self).setUp()

    def tearDown(self):
        super(TestObjectGroup, self).tearDown()

    def test_init(self):
        """
        Test that an ObjectGroup object can be instantiated.
        """
        object_group = objects.ObjectGroup()

        self.assertIsNone(object_group.object_group)

    def test_invalid_object_group(self):
        """
        Test that a TypeError is raised when an invalid object group value
        is used to construct an ObjectGroup attribute.
        """
        kwargs = {"object_group": []}
        self.assertRaisesRegex(
            TypeError,
            "The object group must be a string.",
            objects.ObjectGroup,
            **kwargs
        )

        args = (
            objects.ObjectGroup(),
            "object_group",
            []
        )
        self.assertRaisesRegex(
            TypeError,
            "The object group must be a string.",
            setattr,
            *args
        )

    def test_repr(self):
        """
        Test that repr can be applied to an ObjectGroup attribute.
        """
        object_group = objects.ObjectGroup(object_group="Group1")

        expected = "ObjectGroup({})".format(
            "object_group='{}'".format("Group1")
        )
        observed = repr(object_group)

        self.assertEqual(expected, observed)

    def test_str(self):
        """
        Test that str can be applied to an ObjectGroup attribute.
        """
        object_group = objects.ObjectGroup(object_group="Group1")

        expected = str(
            {
                "object_group": "Group1"
            }
        )
        observed = str(object_group)

        self.assertEqual(expected, observed)

    def test_comparison_on_equal(self):
        """
        Test that the equality/inequality operators return True/False when
        comparing two ObjectGroup attributes with the same
        data.
        """
        a = objects.ObjectGroup()
        b = objects.ObjectGroup()

        self.assertTrue(a == b)
        self.assertTrue(b == a)
        self.assertFalse(a != b)
        self.assertFalse(b != a)

        a = objects.ObjectGroup(object_group="Group1")
        b = objects.ObjectGroup(object_group="Group1")

        self.assertTrue(a == b)
        self.assertTrue(b == a)
        self.assertFalse(a != b)
        self.assertFalse(b != a)

    def test_comparison_on_different_object_groups(self):
        """
        Test that the equality/inequality operators return False/True when
        comparing two ObjectGroup attributes with different object groups.
        """
        a = objects.ObjectGroup(object_group="a")
        b = objects.ObjectGroup(object_group="b")

        self.assertFalse(a == b)
        self.assertFalse(b == a)
        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_comparison_on_type_mismatch(self):
        """
        Test that the equality/inequality operators return False/True when
        comparing an ObjectGroup attribute to a non ObjectGroup attribute.
        """
        a = objects.ObjectGroup()
        b = "invalid"

        self.assertFalse(a == b)
        self.assertFalse(b == a)
        self.assertTrue(a != b)
        self.assertTrue(b != a)

    def test_save(self):
        """
        Test that an ObjectGroup attribute can be saved using SQLAlchemy. This
        test will add an attribute instance to the database, verify that no
        exceptions are thrown, and check that its ID was set.
        """
        object_group = objects.ObjectGroup(object_group="Group1")

        engine = sqlalchemy.create_engine("sqlite:///:memory:", echo=True)
        sqltypes.Base.metadata.create_all(engine)

        session = sqlalchemy.orm.sessionmaker(bind=engine)()
        session.add(object_group)
        session.commit()

        self.assertIsNotNone(object_group.id)

    def test_get(self):
        """
        Test that an ObjectGroup attribute can be saved and then retrieved
        using SQLAlchemy. This test adds the attribute to the database and then
        retrieves it by ID and verifies its values.
        """
        object_group = objects.ObjectGroup(object_group="Group1")

        engine = sqlalchemy.create_engine("sqlite:///:memory:", echo=True)
        sqltypes.Base.metadata.create_all(engine)

        session = sqlalchemy.orm.sessionmaker(bind=engine)()
        session.add(object_group)
        session.commit()

        # Grab the ID now before making a new session to avoid a Detached error
        # See http://sqlalche.me/e/bhk3 for more info.
        object_group_id = object_group.id

        session = sqlalchemy.orm.sessionmaker(bind=engine)()
        retrieved_group = session.query(
            objects.ObjectGroup
        ).filter(
            objects.ObjectGroup.id == object_group_id
        ).one()
        session.commit()

        self.assertEqual("Group1", retrieved_group.object_group)



# Copyright (c) 2014 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six

from kmip.core import enums
from kmip.core.enums import HashingAlgorithm as HashingAlgorithmEnum
from kmip.core.enums import KeyFormatType as KeyFormatTypeEnum
from kmip.core.enums import Tags

from kmip.core import exceptions

from kmip.core.misc import KeyFormatType

from kmip.core import primitives
from kmip.core.primitives import Boolean
from kmip.core.primitives import ByteString
from kmip.core.primitives import Enumeration
from kmip.core.primitives import Integer
from kmip.core.primitives import Struct
from kmip.core.primitives import TextString

from kmip.core import utils
from kmip.core.utils import BytearrayStream

from enum import Enum


# 3.1
class UniqueIdentifier(TextString):

    def __init__(self, value=None, tag=Tags.UNIQUE_IDENTIFIER):
        super(UniqueIdentifier, self).__init__(value, tag)


class PrivateKeyUniqueIdentifier(UniqueIdentifier):

    def __init__(self, value=None):
        super(PrivateKeyUniqueIdentifier, self).__init__(
            value, Tags.PRIVATE_KEY_UNIQUE_IDENTIFIER)


class PublicKeyUniqueIdentifier(UniqueIdentifier):

    def __init__(self, value=None):
        super(PublicKeyUniqueIdentifier, self).__init__(
            value, Tags.PUBLIC_KEY_UNIQUE_IDENTIFIER)


# 3.2
class Name(Struct):

    class NameValue(TextString):

        def __init__(self, value=None):
            super(Name.NameValue, self).__init__(value, Tags.NAME_VALUE)

        def __eq__(self, other):
            if isinstance(other, Name.NameValue):
                if self.value == other.value:
                    return True
                else:
                    return False
            else:
                return NotImplemented

        def __repr__(self):
            return "{0}(value={1})".format(
                    type(self).__name__, repr(self.value))

        def __str__(self):
            return "{0}".format(self.value)

    class NameType(Enumeration):

        def __init__(self, value=None):
            super(Name.NameType, self).__init__(
                enums.NameType, value, Tags.NAME_TYPE)

        def __eq__(self, other):
            if isinstance(other, Name.NameType):
                if self.value == other.value:
                    return True
                else:
                    return False
            else:
                return NotImplemented

        def __repr__(self):
            return "{0}(value={1})".format(
                    type(self).__name__, repr(self.value))

        def __str__(self):
            return "{0}".format(self.value)

    def __init__(self, name_value=None, name_type=None):
        super(Name, self).__init__(tag=Tags.NAME)
        self.name_value = name_value
        self.name_type = name_type
        self.validate()

    def read(self, istream, kmip_version=enums.KMIPVersion.KMIP_1_0):
        super(Name, self).read(istream, kmip_version=kmip_version)
        tstream = BytearrayStream(istream.read(self.length))

        # Read the value and type of the name
        self.name_value = Name.NameValue()
        self.name_type = Name.NameType()
        self.name_value.read(tstream, kmip_version=kmip_version)
        self.name_type.read(tstream, kmip_version=kmip_version)

        self.is_oversized(tstream)

    def write(self, ostream, kmip_version=enums.KMIPVersion.KMIP_1_0):
        tstream = BytearrayStream()

        # Write the value and type of the name
        self.name_value.write(tstream, kmip_version=kmip_version)
        self.name_type.write(tstream, kmip_version=kmip_version)

        # Write the length and value of the template attribute
        self.length = tstream.length()
        super(Name, self).write(ostream, kmip_version=kmip_version)
        ostream.write(tstream.buffer)

    def validate(self):
        self.__validate()

    def __validate(self):
        name = Name.__name__
        msg = exceptions.ErrorStrings.BAD_EXP_RECV
        if self.name_value and \
                not isinstance(self.name_value, Name.NameValue) and \
                not isinstance(self.name_value, str):
            member = 'name_value'
            raise TypeError(msg.format('{0}.{1}'.format(name, member),
                                       'name_value', type(Name.NameValue),
                                       type(self.name_value)))
        if self.name_type and \
                not isinstance(self.name_type, Name.NameType) and \
                not isinstance(self.name_type, str):
            member = 'name_type'
            raise TypeError(msg.format('{0}.{1}'.format(name, member),
                                       'name_type', type(Name.NameType),
                                       type(self.name_type)))

    @classmethod
    def create(cls, name_value, name_type):
        '''
            Returns a Name object, populated with the given value and type
        '''
        if isinstance(name_value, Name.NameValue):
            value = name_value
        elif isinstance(name_value, str):
            value = cls.NameValue(name_value)
        else:
            name = 'Name'
            msg = exceptions.ErrorStrings.BAD_EXP_RECV
            member = 'name_value'
            raise TypeError(msg.format('{0}.{1}'.format(name, member),
                                       'name_value', type(Name.NameValue),
                                       type(name_value)))

        if isinstance(name_type, Name.NameType):
            n_type = name_type
        elif isinstance(name_type, Enum):
            n_type = cls.NameType(name_type)
        else:
            name = 'Name'
            msg = exceptions.ErrorStrings.BAD_EXP_RECV
            member = 'name_type'
            raise TypeError(msg.format('{0}.{1}'.format(name, member),
                                       'name_type', type(Name.NameType),
                                       type(name_type)))

        return Name(name_value=value,
                    name_type=n_type)

    def __repr__(self):
        return "{0}(type={1},value={2})".format(
                type(self).__name__,
                repr(self.name_type),
                repr(self.name_value))

    def __str__(self):
        return "{0}".format(self.name_value.value)

    def __eq__(self, other):
        if isinstance(other, Name):
            if self.name_value == other.name_value and \
                        self.name_type == other.name_type:
                return True
            else:
                return False
        else:
            return NotImplemented

    def __ne__(self, other):
        return not self.__eq__(other)


# 3.3
class ObjectType(Enumeration):

    def __init__(self, value=None):
        super(ObjectType, self).__init__(
            enums.ObjectType, value, Tags.OBJECT_TYPE)


# 3.4
class CryptographicAlgorithm(Enumeration):

    def __init__(self, value=None):
        super(CryptographicAlgorithm, self).__init__(
            enums.CryptographicAlgorithm, value, Tags.CRYPTOGRAPHIC_ALGORITHM)


# 3.5
class CryptographicLength(Integer):

    def __init__(self, value=None):
        super(CryptographicLength, self).__init__(
            value, Tags.CRYPTOGRAPHIC_LENGTH)


# 3.6
class HashingAlgorithm(Enumeration):
    """
    An encodeable wrapper for the HashingAlgorithm enumeration.

    Used to specify the algorithm used to compute the Digest of a Managed
    Object. See Sections 3.17 and 9.1.3.2.16 of the KMIP v1.1 specification
    for more information.
    """

    def __init__(self, value=HashingAlgorithmEnum.SHA_256):
        """
        Construct a HashingAlgorithm object.

        Args:
            value (HashingAlgorithm): A HashingAlgorithm enumeration value,
                (e.g., HashingAlgorithm.MD5). Optional, defaults to
                HashingAlgorithm.SHA_256.
        """
        super(HashingAlgorithm, self).__init__(
            enums.HashingAlgorithm, value, Tags.HASHING_ALGORITHM)


class CryptographicParameters(Struct):
    """
    A set of values for cryptographic operations.

    A structure containing optional fields describing certain cryptographic
    parameters to be used when performing cryptographic operations with the
    associated KMIP object.
    """

    def __init__(self,
                 block_cipher_mode=None,
                 padding_method=None,
                 hashing_algorithm=None,
                 key_role_type=None,
                 digital_signature_algorithm=None,
                 cryptographic_algorithm=None,
                 random_iv=None,
                 iv_length=None,
                 tag_length=None,
                 fixed_field_length=None,
                 invocation_field_length=None,
                 counter_length=None,
                 initial_counter_value=None):
        super(CryptographicParameters, self).__init__(
            tag=Tags.CRYPTOGRAPHIC_PARAMETERS)

        self._block_cipher_mode = None
        self._padding_method = None
        self._hashing_algorithm = None
        self._key_role_type = None
        self._digital_signature_algorithm = None
        self._cryptographic_algorithm = None
        self._random_iv = None
        self._iv_length = None
        self._tag_length = None
        self._fixed_field_length = None
        self._invocation_field_length = None
        self._counter_length = None
        self._initial_counter_value = None

        self.block_cipher_mode = block_cipher_mode
        self.padding_method = padding_method
        self.hashing_algorithm = hashing_algorithm
        self.key_role_type = key_role_type
        self.digital_signature_algorithm = digital_signature_algorithm
        self.cryptographic_algorithm = cryptographic_algorithm
        self.random_iv = random_iv
        self.iv_length = iv_length
        self.tag_length = tag_length
        self.fixed_field_length = fixed_field_length
        self.invocation_field_length = invocation_field_length
        self.counter_length = counter_length
        self.initial_counter_value = initial_counter_value

    @property
    def block_cipher_mode(self):
        if self._block_cipher_mode:
            return self._block_cipher_mode.value
        else:
            return None

    @block_cipher_mode.setter
    def block_cipher_mode(self, value):
        if value is None:
            self._block_cipher_mode = None
        elif isinstance(value, enums.BlockCipherMode):
            self._block_cipher_mode = Enumeration(
                enums.BlockCipherMode,
                value=value,
                tag=Tags.BLOCK_CIPHER_MODE
            )
        else:
            raise TypeError(
                "block cipher mode must be a BlockCipherMode enumeration"
            )

    @property
    def padding_method(self):
        if self._padding_method:
            return self._padding_method.value
        else:
            return None

    @padding_method.setter
    def padding_method(self, value):
        if value is None:
            self._padding_method = None
        elif isinstance(value, enums.PaddingMethod):
            self._padding_method = Enumeration(
                enums.PaddingMethod,
                value=value,
                tag=Tags.PADDING_METHOD
            )
        else:
            raise TypeError(
                "padding method must be a PaddingMethod enumeration"
            )

    @property
    def hashing_algorithm(self):
        if self._hashing_algorithm:
            return self._hashing_algorithm.value
        else:
            return None

    @hashing_algorithm.setter
    def hashing_algorithm(self, value):
        if value is None:
            self._hashing_algorithm = None
        elif isinstance(value, enums.HashingAlgorithm):
            self._hashing_algorithm = Enumeration(
                enums.HashingAlgorithm,
                value=value,
                tag=Tags.HASHING_ALGORITHM
            )
        else:
            raise TypeError(
                "hashing algorithm must be a HashingAlgorithm enumeration"
            )

    @property
    def key_role_type(self):
        if self._key_role_type:
            return self._key_role_type.value
        else:
            return None

    @key_role_type.setter
    def key_role_type(self, value):
        if value is None:
            self._key_role_type = None
        elif isinstance(value, enums.KeyRoleType):
            self._key_role_type = Enumeration(
                enums.KeyRoleType,
                value=value,
                tag=Tags.KEY_ROLE_TYPE
            )
        else:
            raise TypeError(
                "key role type must be a KeyRoleType enumeration"
            )

    @property
    def digital_signature_algorithm(self):
        if self._digital_signature_algorithm:
            return self._digital_signature_algorithm.value
        else:
            return None

    @digital_signature_algorithm.setter
    def digital_signature_algorithm(self, value):
        if value is None:
            self._digital_signature_algorithm = None
        elif isinstance(value, enums.DigitalSignatureAlgorithm):
            self._digital_signature_algorithm = Enumeration(
                enums.DigitalSignatureAlgorithm,
                value=value,
                tag=Tags.DIGITAL_SIGNATURE_ALGORITHM
            )
        else:
            raise TypeError(
                "digital signature algorithm must be a "
                "DigitalSignatureAlgorithm enumeration"
            )

    @property
    def cryptographic_algorithm(self):
        if self._cryptographic_algorithm:
            return self._cryptographic_algorithm.value
        else:
            return None

    @cryptographic_algorithm.setter
    def cryptographic_algorithm(self, value):
        if value is None:
            self._cryptographic_algorithm = None
        elif isinstance(value, enums.CryptographicAlgorithm):
            self._cryptographic_algorithm = Enumeration(
                enums.CryptographicAlgorithm,
                value=value,
                tag=Tags.CRYPTOGRAPHIC_ALGORITHM
            )
        else:
            raise TypeError(
                "cryptographic algorithm must be a CryptographicAlgorithm "
                "enumeration"
            )

    @property
    def random_iv(self):
        if self._random_iv:
            return self._random_iv.value
        else:
            return None

    @random_iv.setter
    def random_iv(self, value):
        if value is None:
            self._random_iv = None
        elif isinstance(value, bool):
            self._random_iv = Boolean(
                value=value,
                tag=Tags.RANDOM_IV
            )
        else:
            raise TypeError("random iv must be a boolean")

    @property
    def iv_length(self):
        if self._iv_length:
            return self._iv_length.value
        else:
            return None

    @iv_length.setter
    def iv_length(self, value):
        if value is None:
            self._iv_length = None
        elif isinstance(value, six.integer_types):
            self._iv_length = Integer(
                value=value,
                tag=Tags.IV_LENGTH
            )
        else:
            raise TypeError("iv length must be an integer")

    @property
    def tag_length(self):
        if self._tag_length:
            return self._tag_length.value
        else:
            return None

    @tag_length.setter
    def tag_length(self, value):
        if value is None:
            self._tag_length = None
        elif isinstance(value, six.integer_types):
            self._tag_length = Integer(
                value=value,
                tag=Tags.TAG_LENGTH
            )
        else:
            raise TypeError("tag length must be an integer")

    @property
    def fixed_field_length(self):
        if self._fixed_field_length:
            return self._fixed_field_length.value
        else:
            return None

    @fixed_field_length.setter
    def fixed_field_length(self, value):
        if value is None:
            self._fixed_field_length = None
        elif isinstance(value, six.integer_types):
            self._fixed_field_length = Integer(
                value=value,
                tag=Tags.FIXED_FIELD_LENGTH
            )
        else:
            raise TypeError("fixed field length must be an integer")

    @property
    def invocation_field_length(self):
        if self._invocation_field_length:
            return self._invocation_field_length.value
        else:
            return None

    @invocation_field_length.setter
    def invocation_field_length(self, value):
        if value is None:
            self._invocation_field_length = None
        elif isinstance(value, six.integer_types):
            self._invocation_field_length = Integer(
                value=value,
                tag=Tags.INVOCATION_FIELD_LENGTH
            )
        else:
            raise TypeError("invocation field length must be an integer")

    @property
    def counter_length(self):
        if self._counter_length:
            return self._counter_length.value
        else:
            return None

    @counter_length.setter
    def counter_length(self, value):
        if value is None:
            self._counter_length = None
        elif isinstance(value, six.integer_types):
            self._counter_length = Integer(
                value=value,
                tag=Tags.COUNTER_LENGTH
            )
        else:
            raise TypeError("counter length must be an integer")

    @property
    def initial_counter_value(self):
        if self._initial_counter_value:
            return self._initial_counter_value.value
        else:
            return None

    @initial_counter_value.setter
    def initial_counter_value(self, value):
        if value is None:
            self._initial_counter_value = None
        elif isinstance(value, six.integer_types):
            self._initial_counter_value = Integer(
                value=value,
                tag=Tags.INITIAL_COUNTER_VALUE
            )
        else:
            raise TypeError("initial counter value must be an integer")

    def read(self, istream, kmip_version=enums.KMIPVersion.KMIP_1_0):
        super(CryptographicParameters, self).read(
            istream,
            kmip_version=kmip_version
        )
        tstream = BytearrayStream(istream.read(self.length))

        if self.is_tag_next(Tags.BLOCK_CIPHER_MODE, tstream):
            self._block_cipher_mode = Enumeration(
                enums.BlockCipherMode,
                tag=Tags.BLOCK_CIPHER_MODE
            )
            self._block_cipher_mode.read(tstream, kmip_version=kmip_version)

        if self.is_tag_next(Tags.PADDING_METHOD, tstream):
            self._padding_method = Enumeration(
                enums.PaddingMethod,
                tag=Tags.PADDING_METHOD
            )
            self._padding_method.read(tstream, kmip_version=kmip_version)

        if self.is_tag_next(Tags.HASHING_ALGORITHM, tstream):
            self._hashing_algorithm = Enumeration(
                enums.HashingAlgorithm,
                tag=Tags.HASHING_ALGORITHM
            )
            self._hashing_algorithm.read(tstream, kmip_version=kmip_version)

        if self.is_tag_next(Tags.KEY_ROLE_TYPE, tstream):
            self._key_role_type = Enumeration(
                enums.KeyRoleType,
                tag=Tags.KEY_ROLE_TYPE
            )
            self._key_role_type.read(tstream, kmip_version=kmip_version)

        if self.is_tag_next(Tags.DIGITAL_SIGNATURE_ALGORITHM, tstream):
            self._digital_signature_algorithm = Enumeration(
                enums.DigitalSignatureAlgorithm,
                tag=Tags.DIGITAL_SIGNATURE_ALGORITHM
            )
            self._digital_signature_algorithm.read(
                tstream,
                kmip_version=kmip_version
            )

        if self.is_tag_next(Tags.CRYPTOGRAPHIC_ALGORITHM, tstream):
            self._cryptographic_algorithm = Enumeration(
                enums.CryptographicAlgorithm,
                tag=Tags.CRYPTOGRAPHIC_ALGORITHM
            )
            self._cryptographic_algorithm.read(
                tstream,
                kmip_version=kmip_version
            )

        if self.is_tag_next(Tags.RANDOM_IV, tstream):
            self._random_iv = Boolean(tag=Tags.RANDOM_IV)
            self._random_iv.read(tstream, kmip_version=kmip_version)

        if self.is_tag_next(Tags.IV_LENGTH, tstream):
            self._iv_length = Integer(tag=Tags.IV_LENGTH)
            self._iv_length.read(tstream, kmip_version=kmip_version)

        if self.is_tag_next(Tags.TAG_LENGTH, tstream):
            self._tag_length = Integer(tag=Tags.TAG_LENGTH)
            self._tag_length.read(tstream, kmip_version=kmip_version)

        if self.is_tag_next(Tags.FIXED_FIELD_LENGTH, tstream):
            self._fixed_field_length = Integer(tag=Tags.FIXED_FIELD_LENGTH)
            self._fixed_field_length.read(tstream, kmip_version=kmip_version)

        if self.is_tag_next(Tags.INVOCATION_FIELD_LENGTH, tstream):
            self._invocation_field_length = Integer(
                tag=Tags.INVOCATION_FIELD_LENGTH
            )
            self._invocation_field_length.read(
                tstream,
                kmip_version=kmip_version
            )

        if self.is_tag_next(Tags.COUNTER_LENGTH, tstream):
            self._counter_length = Integer(tag=Tags.COUNTER_LENGTH)
            self._counter_length.read(tstream, kmip_version=kmip_version)

        if self.is_tag_next(Tags.INITIAL_COUNTER_VALUE, tstream):
            self._initial_counter_value = Integer(
                tag=Tags.INITIAL_COUNTER_VALUE
            )
            self._initial_counter_value.read(
                tstream,
                kmip_version=kmip_version
            )

        self.is_oversized(tstream)

    def write(self, ostream, kmip_version=enums.KMIPVersion.KMIP_1_0):
        tstream = BytearrayStream()

        if self._block_cipher_mode:
            self._block_cipher_mode.write(tstream, kmip_version=kmip_version)
        if self._padding_method:
            self._padding_method.write(tstream, kmip_version=kmip_version)
        if self._hashing_algorithm:
            self._hashing_algorithm.write(tstream, kmip_version=kmip_version)
        if self._key_role_type:
            self._key_role_type.write(tstream, kmip_version=kmip_version)
        if self._digital_signature_algorithm:
            self._digital_signature_algorithm.write(
                tstream,
                kmip_version=kmip_version
            )
        if self._cryptographic_algorithm:
            self._cryptographic_algorithm.write(
                tstream,
                kmip_version=kmip_version
            )
        if self._random_iv:
            self._random_iv.write(tstream, kmip_version=kmip_version)
        if self._iv_length:
            self._iv_length.write(tstream, kmip_version=kmip_version)
        if self._tag_length:
            self._tag_length.write(tstream, kmip_version=kmip_version)
        if self._fixed_field_length:
            self._fixed_field_length.write(tstream, kmip_version=kmip_version)
        if self._invocation_field_length:
            self._invocation_field_length.write(
                tstream,
                kmip_version=kmip_version
            )
        if self._counter_length:
            self._counter_length.write(tstream, kmip_version=kmip_version)
        if self._initial_counter_value:
            self._initial_counter_value.write(
                tstream,
                kmip_version=kmip_version
            )

        self.length = tstream.length()
        super(CryptographicParameters, self).write(
            ostream,
            kmip_version=kmip_version
        )
        ostream.write(tstream.buffer)

    def __eq__(self, other):
        if isinstance(other, CryptographicParameters):
            if self.block_cipher_mode != other.block_cipher_mode:
                return False
            elif self.padding_method != other.padding_method:
                return False
            elif self.hashing_algorithm != other.hashing_algorithm:
                return False
            elif self.key_role_type != other.key_role_type:
                return False
            elif self.digital_signature_algorithm \
                    != other.digital_signature_algorithm:
                return False
            elif self.cryptographic_algorithm != other.cryptographic_algorithm:
                return False
            elif self.random_iv != other.random_iv:
                return False
            elif self.iv_length != other.iv_length:
                return False
            elif self.tag_length != other.tag_length:
                return False
            elif self.fixed_field_length != other.fixed_field_length:
                return False
            elif self.invocation_field_length != other.invocation_field_length:
                return False
            elif self.counter_length != other.counter_length:
                return False
            elif self.initial_counter_value != other.initial_counter_value:
                return False
            else:
                return True

    def __ne__(self, other):
        if isinstance(other, CryptographicParameters):
            return not self == other
        else:
            return NotImplemented

    def __repr__(self):
        args = ", ".join([
            "block_cipher_mode={0}".format(self.block_cipher_mode),
            "padding_method={0}".format(self.padding_method),
            "hashing_algorithm={0}".format(self.hashing_algorithm),
            "key_role_type={0}".format(self.key_role_type),
            "digital_signature_algorithm={0}".format(
                self.digital_signature_algorithm
            ),
            "cryptographic_algorithm={0}".format(
                self.cryptographic_algorithm
            ),
            "random_iv={0}".format(self.random_iv),
            "iv_length={0}".format(self.iv_length),
            "tag_length={0}".format(self.tag_length),
            "fixed_field_length={0}".format(self.fixed_field_length),
            "invocation_field_length={0}".format(
                self.invocation_field_length
            ),
            "counter_length={0}".format(self.counter_length),
            "initial_counter_value={0}".format(self.initial_counter_value)
        ])
        return "CryptographicParameters({0})".format(args)

    def __str__(self):
        return str({
            'block_cipher_mode': self.block_cipher_mode,
            'padding_method': self.padding_method,
            'hashing_algorithm': self.hashing_algorithm,
            'key_role_type': self.key_role_type,
            'digital_signature_algorithm': self.digital_signature_algorithm,
            'cryptographic_algorithm': self.cryptographic_algorithm,
            'random_iv': self.random_iv,
            'iv_length': self.iv_length,
            'tag_length': self.tag_length,
            'fixed_field_length': self.fixed_field_length,
            'invocation_field_length': self.invocation_field_length,
            'counter_length': self.counter_length,
            'initial_counter_value': self.initial_counter_value
        })


class CertificateType(Enumeration):
    """
    An encodeable wrapper for the CertificateType enumeration.

    Used to specify the type of the encoded bytes of a Certificate Managed
    Object. See Sections 2.2.1 and 3.8 of the KMIP v1.1 specification for more
    information.
    """

    def __init__(self, value=enums.CertificateType.X_509):
        """
        Construct a CertificateType object.

        Args:
            value (CertificateType): A CertificateType enumeration
                value, (e.g., CertificateType.PGP). Optional, defaults to
                CertificateType.X_509.
        """
        super(CertificateType, self).__init__(
            enums.CertificateType, value, Tags.CERTIFICATE_TYPE)


class DigestValue(ByteString):
    """
    A byte string representing the hash value of a Digest.

    Used to hold the bytes of the digest hash value. Automatically generated
    by the KMIP server, the value is empty if the server does not have access
    to the value or encoding of the related Managed Object. See Section 3.17
    of the KMIP 1.1 specification for more information.

    Attributes:
        value: The bytes of the hash.
    """

    def __init__(self, value=b''):
        """
        Construct a DigestValue object.

        Args:
            value (bytes): The bytes of the hash. Optional, defaults to
                the empty byte string.
        """
        super(DigestValue, self).__init__(value, Tags.DIGEST_VALUE)


class Digest(Struct):
    """
    A structure storing a hash digest of a Managed Object.

    Digests may be calculated for keys, secret data objects, certificates, and
    opaque data objects and are generated when the object is created or
    registered with the KMIP server. See Section 3.17 of the KMIP 1.1
    specification for more information.

    Attributes:
        hashing_algorithm: The algorithm used to compute the hash digest.
        digest_value: The bytes representing the hash digest value.
        key_format_type: The type of the key the hash was generated for.
    """

    def __init__(self,
                 hashing_algorithm=None,
                 digest_value=None,
                 key_format_type=None):
        """
        Construct a Digest object.

        Args:
            hashing_algorithm (HashingAlgorithm): The hash algorithm used to
                compute the value of the digest. Optional, defaults to None.
            digest_value (DigestValue): The byte string representing the
                value of the hash digest. Optional, defaults to None.
            key_format_type (KeyFormatType): The format type of the key the
                hash was computed for, if the object in question is a key.
                Optional, defaults to None.
        """
        super(Digest, self).__init__(Tags.DIGEST)

        if hashing_algorithm is None:
            self.hashing_algorithm = HashingAlgorithm()
        else:
            self.hashing_algorithm = hashing_algorithm

        if digest_value is None:
            self.digest_value = DigestValue()
        else:
            self.digest_value = digest_value

        if key_format_type is None:
            self.key_format_type = KeyFormatType()
        else:
            self.key_format_type = key_format_type

        self.validate()

    def read(self, istream, kmip_version=enums.KMIPVersion.KMIP_1_0):
        """
        Read the data encoding the Digest object and decode it into its
        constituent parts.

        Args:
            istream (Stream): A data stream containing encoded object data,
                supporting a read method; usually a BytearrayStream object.
            kmip_version (KMIPVersion): An enumeration defining the KMIP
                version with which the object will be decoded. Optional,
                defaults to KMIP 1.0.
        """
        super(Digest, self).read(istream, kmip_version=kmip_version)
        tstream = BytearrayStream(istream.read(self.length))

        self.hashing_algorithm.read(tstream, kmip_version=kmip_version)
        self.digest_value.read(tstream, kmip_version=kmip_version)
        self.key_format_type.read(tstream, kmip_version=kmip_version)

        self.is_oversized(tstream)
        self.validate()

    def write(self, ostream, kmip_version=enums.KMIPVersion.KMIP_1_0):
        """
        Write the data encoding the Digest object to a stream.

        Args:
            ostream (Stream): A data stream in which to encode object data,
                supporting a write method; usually a BytearrayStream object.
            kmip_version (KMIPVersion): An enumeration defining the KMIP
                version with which the object will be encoded. Optional,
                defaults to KMIP 1.0.
        """
        tstream = BytearrayStream()

        self.hashing_algorithm.write(tstream, kmip_version=kmip_version)
        self.digest_value.write(tstream, kmip_version=kmip_version)
        self.key_format_type.write(tstream, kmip_version=kmip_version)

        self.length = tstream.length()
        super(Digest, self).write(ostream, kmip_version=kmip_version)
        ostream.write(tstream.buffer)

    def validate(self):
        """
        Error check the attributes of the Digest object.
        """
        self.__validate()

    def __validate(self):
        # TODO (peter-hamilton) Add checks comparing the length of the digest
        # value against the standard length for the stated hashing algorithm.
        if not isinstance(self.hashing_algorithm, HashingAlgorithm):
            msg = "invalid hashing algorithm"
            msg += "; expected {0}, received {1}".format(
                HashingAlgorithm, self.hashing_algorithm)
            raise TypeError(msg)

        if not isinstance(self.digest_value, DigestValue):
            msg = "invalid digest value"
            msg += "; expected {0}, received {1}".format(
                DigestValue, self.digest_value)
            raise TypeError(msg)

        if not isinstance(self.key_format_type, KeyFormatType):
            msg = "invalid key format type"
            msg += "; expected {0}, received {1}".format(
                KeyFormatType, self.key_format_type)
            raise TypeError(msg)

    def __eq__(self, other):
        if isinstance(other, Digest):
            if self.hashing_algorithm != other.hashing_algorithm:
                return False
            elif self.digest_value != other.digest_value:
                return False
            elif self.key_format_type != other.key_format_type:
                return False
            else:
                return True
        else:
            return NotImplemented

    def __ne__(self, other):
        if isinstance(other, Digest):
            return not (self == other)
        else:
            return NotImplemented

    def __repr__(self):
        hashing_algorithm = "hashing_algorithm={0}".format(
            repr(self.hashing_algorithm))
        digest_value = "digest_value={0}".format(
            repr(self.digest_value))
        key_format_type = "key_format_type={0}".format(
            repr(self.key_format_type))

        return "Digest({0}, {1}, {2})".format(
            hashing_algorithm, digest_value, key_format_type)

    def __str__(self):
        return str(self.digest_value)

    @classmethod
    def create(cls,
               hashing_algorithm=HashingAlgorithmEnum.SHA_256,
               digest_value=b'',
               key_format_type=KeyFormatTypeEnum.RAW):
        """
        Construct a Digest object from provided digest values.

        Args:
            hashing_algorithm (HashingAlgorithm): An enumeration representing
                the hash algorithm used to compute the digest. Optional,
                defaults to HashingAlgorithm.SHA_256.
            digest_value (byte string): The bytes of the digest hash. Optional,
                defaults to the empty byte string.
            key_format_type (KeyFormatType): An enumeration representing the
                format of the key corresponding to the digest. Optional,
                defaults to KeyFormatType.RAW.

        Returns:
            Digest: The newly created Digest.

        Example:
            >>> x = Digest.create(HashingAlgorithm.MD5, b'\x00',
            ... KeyFormatType.RAW)
            >>> x.hashing_algorithm
            HashingAlgorithm(value=HashingAlgorithm.MD5)
            >>> x.digest_value
            DigestValue(value=bytearray(b'\x00'))
            >>> x.key_format_type
            KeyFormatType(value=KeyFormatType.RAW)
        """
        algorithm = HashingAlgorithm(hashing_algorithm)
        value = DigestValue(bytearray(digest_value))
        format_type = KeyFormatType(key_format_type)

        return Digest(hashing_algorithm=algorithm,
                      digest_value=value,
                      key_format_type=format_type)


# 3.18
class OperationPolicyName(TextString):

    def __init__(self, value=None):
        super(OperationPolicyName, self).__init__(
            value, Tags.OPERATION_POLICY_NAME)


# 3.19
class CryptographicUsageMask(Integer):

    ENUM_TYPE = enums.CryptographicUsageMask

    def __init__(self, value=None):
        super(CryptographicUsageMask, self).__init__(
            value, Tags.CRYPTOGRAPHIC_USAGE_MASK)


class State(Enumeration):

    def __init__(self, value=None):
        super(State, self).__init__(enums.State, value, Tags.STATE)


class ApplicationSpecificInformation(primitives.Struct):
    """
    A structure used to store data specific to the applications that use a
    Managed Object.

    An attribute of Managed Objects, it may be specified during the creation or
    modification of any server Managed Object.

    Attributes:
        application_namespace: The name of a namespace supported by the server.
        application_data: String data relevant to the specified namespace.

    See Section 3.36 of the KMIP v1.1 specification for more information.
    """

    def __init__(self, application_namespace=None, application_data=None):
        """
        Construct an ApplicationSpecificInformation object.

        Args:
            application_namespace (string): The name of a namespace supported
                by the server. Optional, defaults to None. Required for
                read/write.
            application_data (string): String data relevant to the specified
                namespace. Optional, defaults to None. Required for read/write.
        """
        super(ApplicationSpecificInformation, self).__init__(
            enums.Tags.APPLICATION_SPECIFIC_INFORMATION
        )

        self._application_namespace = None
        self._application_data = None

        self.application_namespace = application_namespace
        self.application_data = application_data

    @property
    def application_namespace(self):
        if self._application_namespace:
            return self._application_namespace.value
        return None

    @application_namespace.setter
    def application_namespace(self, value):
        if value is None:
            self._application_namespace = None
        elif isinstance(value, six.string_types):
            self._application_namespace = primitives.TextString(
                value=value,
                tag=enums.Tags.APPLICATION_NAMESPACE
            )
        else:
            raise TypeError("The application namespace must be a string.")

    @property
    def application_data(self):
        if self._application_data:
            return self._application_data.value
        return None

    @application_data.setter
    def application_data(self, value):
        if value is None:
            self._application_data = None
        elif isinstance(value, six.string_types):
            self._application_data = primitives.TextString(
                value=value,
                tag=enums.Tags.APPLICATION_DATA
            )
        else:
            raise TypeError("The application data must be a string.")

    def read(self, input_buffer, kmip_version=enums.KMIPVersion.KMIP_1_0):
        """
        Read the data encoding the ApplicationSpecificInformation attribute
        and decode it.

        Args:
            input_buffer (stream): A data stream containing encoded object
                data, supporting a read method; usually a BytearrayStream
                object.
            kmip_version (KMIPVersion): An enumeration defining the KMIP
                version with which the object will be decoded. Optional,
                defaults to KMIP 1.0.
        """
        super(ApplicationSpecificInformation, self).read(
            input_buffer,
            kmip_version=kmip_version
        )
        local_buffer = utils.BytearrayStream(input_buffer.read(self.length))

        if self.is_tag_next(enums.Tags.APPLICATION_NAMESPACE, local_buffer):
            self._application_namespace = primitives.TextString(
                tag=enums.Tags.APPLICATION_NAMESPACE
            )
            self._application_namespace.read(
                local_buffer,
                kmip_version=kmip_version
            )
        else:
            raise exceptions.InvalidKmipEncoding(
                "The ApplicationSpecificInformation encoding is missing the "
                "ApplicationNamespace field."
            )
        if self.is_tag_next(enums.Tags.APPLICATION_DATA, local_buffer):
            self._application_data = primitives.TextString(
                tag=enums.Tags.APPLICATION_DATA
            )
            self._application_data.read(
                local_buffer,
                kmip_version=kmip_version
            )
        else:
            raise exceptions.InvalidKmipEncoding(
                "The ApplicationSpecificInformation encoding is missing the "
                "ApplicationData field."
            )

        self.is_oversized(local_buffer)

    def write(self, output_buffer, kmip_version=enums.KMIPVersion.KMIP_1_0):
        """
        Write the data encoding the ApplicationSpecificInformation object to a
        buffer.

        Args:
            output_buffer (stream): A data stream in which to encode object
                data, supporting a write method; usually a BytearrayStream
                object.
            kmip_version (KMIPVersion): An enumeration defining the KMIP
                version with which the object will be encoded. Optional,
                defaults to KMIP 1.0.
        """
        local_buffer = utils.BytearrayStream()

        if self._application_namespace:
            self._application_namespace.write(
                local_buffer,
                kmip_version=kmip_version
            )
        else:
            raise exceptions.InvalidField(
                "The ApplicationSpecificInformation object is missing the "
                "ApplicationNamespace field."
            )
        if self._application_data:
            self._application_data.write(
                local_buffer,
                kmip_version=kmip_version
            )
        else:
            raise exceptions.InvalidField(
                "The ApplicationSpecificInformation object is missing the "
                "ApplicationData field."
            )

        self.length = local_buffer.length()
        super(ApplicationSpecificInformation, self).write(
            output_buffer,
            kmip_version=kmip_version
        )
        output_buffer.write(local_buffer.buffer)

    def __repr__(self):
        args = [
            "application_namespace={}".format(
                repr(self.application_namespace)
            ),
            "application_data={}".format(repr(self.application_data))
        ]
        return "ApplicationSpecificInformation({})".format(", ".join(args))

    def __str__(self):
        value = ", ".join(
            [
                '"application_namespace": "{}"'.format(
                    self.application_namespace
                ),
                '"application_data": "{}"'.format(self.application_data)
            ]
        )
        return "{" + value + "}"

    def __eq__(self, other):
        if isinstance(other, ApplicationSpecificInformation):
            if self.application_namespace != other.application_namespace:
                return False
            if self.application_data != other.application_data:
                return False
            return True
        else:
            return NotImplemented

    def __ne__(self, other):
        if isinstance(other, ApplicationSpecificInformation):
            return not self.__eq__(other)
        else:
            return NotImplemented


# 3.37
class ContactInformation(TextString):

    def __init__(self, value=None):
        super(ContactInformation, self).__init__(
            value, Tags.CONTACT_INFORMATION)


# 3.39
# TODO (peter-hamilton) A CustomAttribute TextString is not sufficient to
# TODO (peter-hamilton) cover all potential custom attributes. This is a
# TODO (peter-hamilton) temporary stopgap.
class CustomAttribute(TextString):

    def __init__(self, value=None):
        super(CustomAttribute, self).__init__(value, Tags.ATTRIBUTE_VALUE)


class DerivationParameters(Struct):
    """
    A set of values needed for key or secret derivation.

    A structure containing optional fields describing certain cryptographic
    parameters to be used when performing key or secret derivation operations.
    """

    def __init__(self,
                 cryptographic_parameters=None,
                 initialization_vector=None,
                 derivation_data=None,
                 salt=None,
                 iteration_count=None):
        """
        Construct a DerivationParameters struct.

        Args:
            cryptographic_parameters (CryptographicParameters): A
                CryptographicParameters struct containing the settings for
                the derivation process. Optional, defaults to None. If not
                included, the CryptographicParameters associated with the
                managed object will be used instead.
            initialization_vector (bytes): The IV value to be used with the
                pseudo-random derivation function (PRF). Optional depending
                on the PRF, defaults to None.
            derivation_data (bytes): A data component to be used instead of
                or with a derivation key to derive the new cryptographic
                object. Optional, defaults to None.
            salt (bytes): A salt value required by the PBKDF2 algorithm.
                Optional, defaults to None.
            iteration_count (bytes): An iteration count value required by
                the PBKDF2 algorithm. Optional, defaults to None.
        """
        super(DerivationParameters, self).__init__(
            tag=Tags.DERIVATION_PARAMETERS
        )

        self._cryptographic_parameters = None
        self._initialization_vector = None
        self._derivation_data = None
        self._salt = None
        self._iteration_count = None

        self.cryptographic_parameters = cryptographic_parameters
        self.initialization_vector = initialization_vector
        self.derivation_data = derivation_data
        self.salt = salt
        self.iteration_count = iteration_count

    @property
    def cryptographic_parameters(self):
        return self._cryptographic_parameters

    @cryptographic_parameters.setter
    def cryptographic_parameters(self, value):
        if value is None:
            self._cryptographic_parameters = None
        elif isinstance(value, CryptographicParameters):
            self._cryptographic_parameters = value
        else:
            raise TypeError(
                "cryptographic parameters must be a CryptographicParameters "
                "struct"
            )

    @property
    def initialization_vector(self):
        if self._initialization_vector:
            return self._initialization_vector.value
        else:
            return None

    @initialization_vector.setter
    def initialization_vector(self, value):
        if value is None:
            self._initialization_vector = None
        elif isinstance(value, six.binary_type):
            self._initialization_vector = ByteString(
                value=value,
                tag=enums.Tags.INITIALIZATION_VECTOR
            )
        else:
            raise TypeError("initialization vector must be bytes")

    @property
    def derivation_data(self):
        if self._derivation_data:
            return self._derivation_data.value
        else:
            return None

    @derivation_data.setter
    def derivation_data(self, value):
        if value is None:
            self._derivation_data = None
        elif isinstance(value, six.binary_type):
            self._derivation_data = ByteString(
                value=value,
                tag=enums.Tags.DERIVATION_DATA
            )
        else:
            raise TypeError("derivation data must be bytes")

    @property
    def salt(self):
        if self._salt:
            return self._salt.value
        else:
            return None

    @salt.setter
    def salt(self, value):
        if value is None:
            self._salt = None
        elif isinstance(value, six.binary_type):
            self._salt = ByteString(
                value=value,
                tag=enums.Tags.SALT
            )
        else:
            raise TypeError("salt must be bytes")

    @property
    def iteration_count(self):
        if self._iteration_count:
            return self._iteration_count.value
        else:
            return None

    @iteration_count.setter
    def iteration_count(self, value):
        if value is None:
            self._iteration_count = None
        elif isinstance(value, six.integer_types):
            self._iteration_count = Integer(
                value=value,
                tag=Tags.ITERATION_COUNT
            )
        else:
            raise TypeError("iteration count must be an integer")

    def read(self, input_stream, kmip_version=enums.KMIPVersion.KMIP_1_0):
        """
        Read the data encoding the DerivationParameters struct and decode it
        into its constituent parts.

        Args:
            input_stream (stream): A data stream containing encoded object
                data, supporting a read method; usually a BytearrayStream
                object.
            kmip_version (KMIPVersion): An enumeration defining the KMIP
                version with which the object will be decoded. Optional,
                defaults to KMIP 1.0.
        """
        super(DerivationParameters, self).read(
            input_stream,
            kmip_version=kmip_version
        )
        local_stream = BytearrayStream(input_stream.read(self.length))

        if self.is_tag_next(
                enums.Tags.CRYPTOGRAPHIC_PARAMETERS,
                local_stream
        ):
            self._cryptographic_parameters = CryptographicParameters()
            self._cryptographic_parameters.read(
                local_stream,
                kmip_version=kmip_version
            )

        if self.is_tag_next(enums.Tags.INITIALIZATION_VECTOR, local_stream):
            self._initialization_vector = ByteString(
                tag=enums.Tags.INITIALIZATION_VECTOR
            )
            self._initialization_vector.read(
                local_stream,
                kmip_version=kmip_version
            )

        if self.is_tag_next(enums.Tags.DERIVATION_DATA, local_stream):
            self._derivation_data = ByteString(tag=enums.Tags.DERIVATION_DATA)
            self._derivation_data.read(local_stream, kmip_version=kmip_version)

        if self.is_tag_next(enums.Tags.SALT, local_stream):
            self._salt = ByteString(tag=enums.Tags.SALT)
            self._salt.read(local_stream, kmip_version=kmip_version)

        if self.is_tag_next(Tags.ITERATION_COUNT, local_stream):
            self._iteration_count = Integer(tag=Tags.ITERATION_COUNT)
            self._iteration_count.read(local_stream, kmip_version=kmip_version)

        self.is_oversized(local_stream)

    def write(self, output_stream, kmip_version=enums.KMIPVersion.KMIP_1_0):
        """
        Write the data encoding the DerivationParameters struct to a stream.

        Args:
            output_stream (stream): A data stream in which to encode object
                data, supporting a write method; usually a BytearrayStream
                object.
            kmip_version (KMIPVersion): An enumeration defining the KMIP
                version with which the object will be encoded. Optional,
                defaults to KMIP 1.0.
        """
        local_stream = BytearrayStream()

        if self._cryptographic_parameters:
            self._cryptographic_parameters.write(
                local_stream,
                kmip_version=kmip_version
            )
        if self._initialization_vector:
            self._initialization_vector.write(
                local_stream,
                kmip_version=kmip_version
            )
        if self._derivation_data:
            self._derivation_data.write(
                local_stream,
                kmip_version=kmip_version
            )
        if self._salt:
            self._salt.write(
                local_stream,
                kmip_version=kmip_version
            )
        if self._iteration_count:
            self._iteration_count.write(
                local_stream,
                kmip_version=kmip_version
            )

        self.length = local_stream.length()
        super(DerivationParameters, self).write(
            output_stream,
            kmip_version=kmip_version
        )
        output_stream.write(local_stream.buffer)

    def __eq__(self, other):
        if isinstance(other, DerivationParameters):
            if self.cryptographic_parameters != other.cryptographic_parameters:
                return False
            elif self.initialization_vector != other.initialization_vector:
                return False
            elif self.derivation_data != other.derivation_data:
                return False
            elif self.salt != other.salt:
                return False
            elif self.iteration_count != other.iteration_count:
                return False
            else:
                return True

    def __ne__(self, other):
        if isinstance(other, DerivationParameters):
            return not self == other
        else:
            return NotImplemented

    def __repr__(self):
        args = ", ".join([
            "cryptographic_parameters={0}".format(
                repr(self.cryptographic_parameters)
            ),
            "initialization_vector={0}".format(self.initialization_vector),
            "derivation_data={0}".format(self.derivation_data),
            "salt={0}".format(self.salt),
            "iteration_count={0}".format(
                self.iteration_count
            )
        ])
        return "DerivationParameters({0})".format(args)

    def __str__(self):
        return str({
            'cryptographic_parameters': self.cryptographic_parameters,
            'initialization_vector': self.initialization_vector,
            'derivation_data': self.derivation_data,
            'salt': self.salt,
            'iteration_count': self.iteration_count
        })



# Copyright (c) 2014 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import six

from kmip.core import enums
from kmip.core import objects
from kmip.core import primitives
from kmip.core import utils
from kmip.core.messages.payloads import base


class LocateRequestPayload(base.RequestPayload):
    """
    A request payload for the Locate operation.

    Attributes:
        maximum_items: The maximum number of object identifiers to be returned.
        offset_items: The number of object identifiers to skip when selecting
            the object identifiers to return.
        storage_status_mask: A bit mask specifying which types of stored
            objects should be searched.
        object_group_member: The object group member type for the searched
            objects.
        attributes: The attributes that should be used to filter and match
            objects.
    """

    def __init__(self,
                 maximum_items=None,
                 offset_items=None,
                 storage_status_mask=None,
                 object_group_member=None,
                 attributes=None):
        """
        Construct a Locate request payload structure.

        Args:
            maximum_items (int): An integer specifying the maximum number of
                object identifiers to be returned. Optional, defaults to None.
            offset_items (int): An integer specifying the number of object
                identifiers to skip when selecting the object identifiers to
                return. Optional, defaults to None.
            storage_status_mask (int, list): An integer bit mask or a
                corresponding list of StorageStatusMask enumerations indicating
                which types of stored objects should be searched. Optional,
                defaults to None.
            object_group_member (enum): An ObjectGroupMember enumeration
                specifying the object group member type for the searched
                objects. Optional, defaults to None.
            attributes (list): A list of Attribute structures containing the
                attribute values that should be used to filter and match
                objects. Optional, defaults to None. Required for read/write
                for KMIP 2.0+.
        """
        super(LocateRequestPayload, self).__init__()

        self._maximum_items = None
        self._offset_items = None
        self._storage_status_mask = None
        self._object_group_member = None
        self._attributes = None

        self.maximum_items = maximum_items
        self.offset_items = offset_items
        self.storage_status_mask = storage_status_mask
        self.object_group_member = object_group_member
        self.attributes = attributes

    @property
    def maximum_items(self):
        if self._maximum_items:
            return self._maximum_items.value
        else:
            return None

    @maximum_items.setter
    def maximum_items(self, value):
        if value is None:
            self._maximum_items = None
        elif isinstance(value, six.integer_types):
            self._maximum_items = primitives.Integer(
                value=value,
                tag=enums.Tags.MAXIMUM_ITEMS
            )
        else:
            raise TypeError("Maximum items must be an integer.")

    @property
    def offset_items(self):
        if self._offset_items:
            return self._offset_items.value
        else:
            return None

    @offset_items.setter
    def offset_items(self, value):
        if value is None:
            self._offset_items = None
        elif isinstance(value, six.integer_types):
            self._offset_items = primitives.Integer(
                value=value,
                tag=enums.Tags.OFFSET_ITEMS
            )
        else:
            raise TypeError("Offset items must be an integer.")

    @property
    def storage_status_mask(self):
        if self._storage_status_mask:
            return self._storage_status_mask.value
        else:
            return None

    @storage_status_mask.setter
    def storage_status_mask(self, value):
        if value is None:
            self._storage_status_mask = None
        elif isinstance(value, six.integer_types):
            if enums.is_bit_mask(enums.StorageStatusMask, value):
                self._storage_status_mask = primitives.Integer(
                    value=value,
                    tag=enums.Tags.STORAGE_STATUS_MASK
                )
            else:
                raise TypeError(
                    "Storage status mask must be an integer representing a "
                    "valid StorageStatusMask bit mask."
                )
        else:
            raise TypeError(
                "Storage status mask must be an integer representing a valid "
                "StorageStatusMask bit mask."
            )

    @property
    def object_group_member(self):
        if self._object_group_member:
            return self._object_group_member.value
        else:
            return None

    @object_group_member.setter
    def object_group_member(self, value):
        if value is None:
            self._object_group_member = None
        elif isinstance(value, enums.ObjectGroupMember):
            self._object_group_member = primitives.Enumeration(
                enums.ObjectGroupMember,
                value=value,
                tag=enums.Tags.OBJECT_GROUP_MEMBER
            )
        else:
            raise TypeError(
                "Object group member must be an ObjectGroupMember enumeration."
            )

    @property
    def attributes(self):
        if self._attributes:
            return self._attributes
        return []

    @attributes.setter
    def attributes(self, value):
        if value is None:
            self._attributes = []
        elif isinstance(value, list):
            for v in value:
                if not isinstance(v, objects.Attribute):
                    raise TypeError(
                        "Attributes must be a list of Attribute structures."
                    )
            self._attributes = value
        else:
            raise TypeError(
                "Attributes must be a list of Attribute structures."
            )

    def read(self, input_buffer, kmip_version=enums.KMIPVersion.KMIP_1_0):
        """
        Read the data encoding the Locate request payload and decode it into
        its constituent parts.

        Args:
            input_buffer (stream): A data buffer containing encoded object
                data, supporting a read method.
            kmip_version (KMIPVersion): An enumeration defining the KMIP
                version with which the object will be decoded. Optional,
                defaults to KMIP 1.0.

        Raises:
            InvalidKmipEncoding: Raised if the attributes structure is missing
                from the encoded payload for KMIP 2.0+ encodings.
        """
        super(LocateRequestPayload, self).read(
            input_buffer,
            kmip_version=kmip_version
        )
        local_buffer = utils.BytearrayStream(input_buffer.read(self.length))

        if self.is_tag_next(enums.Tags.MAXIMUM_ITEMS, local_buffer):
            self._maximum_items = primitives.Integer(
                tag=enums.Tags.MAXIMUM_ITEMS
            )
            self._maximum_items.read(
                local_buffer,
                kmip_version=kmip_version
            )

        if self.is_tag_next(enums.Tags.OFFSET_ITEMS, local_buffer):
            self._offset_items = primitives.Integer(
                tag=enums.Tags.OFFSET_ITEMS
            )
            self._offset_items.read(
                local_buffer,
                kmip_version=kmip_version
            )

        if self.is_tag_next(enums.Tags.STORAGE_STATUS_MASK, local_buffer):
            self._storage_status_mask = primitives.Integer(
                tag=enums.Tags.STORAGE_STATUS_MASK
            )
            self._storage_status_mask.read(
                local_buffer,
                kmip_version=kmip_version
            )

        if self.is_tag_next(enums.Tags.OBJECT_GROUP_MEMBER, local_buffer):
            self._object_group_member = primitives.Enumeration(
                enums.ObjectGroupMember,
                tag=enums.Tags.OBJECT_GROUP_MEMBER
            )
            self._object_group_member.read(
                local_buffer,
                kmip_version=kmip_version
            )

        if kmip_version < enums.KMIPVersion.KMIP_2_0:
            while self.is_tag_next(enums.Tags.ATTRIBUTE, local_buffer):
                attribute = objects.Attribute()
                attribute.read(local_buffer, kmip_version=kmip_version)
                self._attributes.append(attribute)
        else:
            if self.is_tag_next(enums.Tags.ATTRIBUTES, local_buffer):
                attributes = objects.Attributes()
                attributes.read(local_buffer, kmip_version=kmip_version)
                # TODO (ph) Add a new utility to avoid using TemplateAttributes
                temp_attr = objects.convert_attributes_to_template_attribute(
                    attributes
                )
                self._attributes = temp_attr.attributes

    def write(self, output_buffer, kmip_version=enums.KMIPVersion.KMIP_1_0):
        """
        Write the data encoding the Locate request payload to a buffer.

        Args:
            output_buffer (stream): A data buffer in which to encode object
                data, supporting a write method.
            kmip_version (KMIPVersion): An enumeration defining the KMIP
                version with which the object will be encoded. Optional,
                defaults to KMIP 1.0.
        """
        local_buffer = utils.BytearrayStream()

        if self._maximum_items:
            self._maximum_items.write(local_buffer, kmip_version=kmip_version)

        if self._offset_items:
            self._offset_items.write(local_buffer, kmip_version=kmip_version)

        if self._storage_status_mask:
            self._storage_status_mask.write(
                local_buffer,
                kmip_version=kmip_version
            )

        if self._object_group_member:
            self._object_group_member.write(
                local_buffer,
                kmip_version=kmip_version
            )

        if kmip_version < enums.KMIPVersion.KMIP_2_0:
            if self._attributes:
                for attribute in self.attributes:
                    attribute.write(
                        local_buffer,
                        kmip_version=kmip_version
                    )
        else:
            if self._attributes:
                # TODO (ph) Add a new utility to avoid using TemplateAttributes
                template_attribute = objects.TemplateAttribute(
                    attributes=self.attributes
                )
                attributes = objects.convert_template_attribute_to_attributes(
                    template_attribute
                )
                attributes.write(local_buffer, kmip_version=kmip_version)

        self.length = local_buffer.length()
        super(LocateRequestPayload, self).write(
            output_buffer,
            kmip_version=kmip_version
        )
        output_buffer.write(local_buffer.buffer)

    def __eq__(self, other):
        if isinstance(other, LocateRequestPayload):
            if self.maximum_items != other.maximum_items:
                return False
            elif self.offset_items != other.offset_items:
                return False
            elif self.storage_status_mask != other.storage_status_mask:
                return False
            elif self.object_group_member != other.object_group_member:
                return False
            elif self.attributes != other.attributes:
                return False
            else:
                return True
        else:
            return NotImplemented

    def __ne__(self, other):
        if isinstance(other, LocateRequestPayload):
            return not (self == other)
        else:
            return NotImplemented

    def __repr__(self):
        args = ", ".join([
            "maximum_items={}".format(self.maximum_items),
            "offset_items={}".format(self.offset_items),
            "storage_status_mask={}".format(self.storage_status_mask),
            "object_group_member={}".format(self.object_group_member),
            "attributes={}".format(
                [repr(attribute) for attribute in self.attributes]
            )
        ])
        return "LocateRequestPayload({})".format(args)

    def __str__(self):
        value = ", ".join(
            [
                '"maximum_items": {}'.format(self.maximum_items),
                '"offset_items": {}'.format(self.offset_items),
                '"storage_status_mask": {}'.format(self.storage_status_mask),
                '"object_group_member": {}'.format(self.object_group_member),
                '"attributes": {}'.format(
                    [str(attribute) for attribute in self.attributes]
                )
            ]
        )
        return '{' + value + '}'


class LocateResponsePayload(base.ResponsePayload):
    """
    A response payload for the Locate operation.

    Attributes:
        located_items: The number of matching objects found by the server.
        unique_identifiers: The object identifiers for the matching objects.
    """

    def __init__(self,
                 located_items=None,
                 unique_identifiers=None):
        """
        Construct a Locate response payload structure.

        Args:
            located_items (int): An integer specifying the number of matching
                objects found by the server. Note that this may not equal the
                number of object identifiers returned in this payload.
                Optional, defaults to None.
            unique_identifiers (list): A list of strings specifying the object
                identifiers for matching objects. Optional, defaults to None.
        """
        super(LocateResponsePayload, self).__init__()

        self._located_items = None
        self._unique_identifiers = None

        self.located_items = located_items
        self.unique_identifiers = unique_identifiers

    @property
    def located_items(self):
        if self._located_items:
            return self._located_items.value
        return None

    @located_items.setter
    def located_items(self, value):
        if value is None:
            self._located_items = None
        elif isinstance(value, six.integer_types):
            self._located_items = primitives.Integer(
                value=value,
                tag=enums.Tags.LOCATED_ITEMS
            )
        else:
            raise TypeError("Located items must be an integer.")

    @property
    def unique_identifiers(self):
        if self._unique_identifiers:
            return [x.value for x in self._unique_identifiers]
        return []

    @unique_identifiers.setter
    def unique_identifiers(self, value):
        if value is None:
            self._unique_identifiers = []
        elif isinstance(value, list):
            self._unique_identifiers = []
            for v in value:
                if not isinstance(v, six.string_types):
                    self._unique_identifiers = []
                    raise TypeError(
                        "Unique identifiers must be a list of strings."
                    )
                self._unique_identifiers.append(
                    primitives.TextString(
                        value=v,
                        tag=enums.Tags.UNIQUE_IDENTIFIER
                    )
                )
        else:
            raise TypeError("Unique identifiers must be a list of strings.")

    def read(self, input_buffer, kmip_version=enums.KMIPVersion.KMIP_1_0):
        """
        Read the data encoding the Locate response payload and decode it
        into its constituent parts.

        Args:
            input_buffer (stream): A data buffer containing encoded object
                data, supporting a read method.
            kmip_version (KMIPVersion): An enumeration defining the KMIP
                version with which the object will be decoded. Optional,
                defaults to KMIP 1.0.
        """
        super(LocateResponsePayload, self).read(
            input_buffer,
            kmip_version=kmip_version
        )
        local_buffer = utils.BytearrayStream(input_buffer.read(self.length))

        if self.is_tag_next(enums.Tags.LOCATED_ITEMS, local_buffer):
            self._located_items = primitives.Integer(
                tag=enums.Tags.LOCATED_ITEMS
            )
            self._located_items.read(
                local_buffer,
                kmip_version=kmip_version
            )

        self._unique_identifiers = []
        while self.is_tag_next(enums.Tags.UNIQUE_IDENTIFIER, local_buffer):
            unique_identifier = primitives.TextString(
                tag=enums.Tags.UNIQUE_IDENTIFIER
            )
            unique_identifier.read(local_buffer, kmip_version=kmip_version)
            self._unique_identifiers.append(unique_identifier)

        self.is_oversized(local_buffer)

    def write(self, output_buffer, kmip_version=enums.KMIPVersion.KMIP_1_0):
        """
        Write the data encoding the Locate response payload to a buffer.

        Args:
            output_buffer (stream): A data buffer in which to encode object
                data, supporting a write method.
            kmip_version (KMIPVersion): An enumeration defining the KMIP
                version with which the object will be encoded. Optional,
                defaults to KMIP 1.0.
        """
        local_buffer = utils.BytearrayStream()

        if self._located_items:
            self._located_items.write(local_buffer, kmip_version=kmip_version)

        if self._unique_identifiers:
            for unique_identifier in self._unique_identifiers:
                unique_identifier.write(
                    local_buffer,
                    kmip_version=kmip_version
                )

        self.length = local_buffer.length()
        super(LocateResponsePayload, self).write(
            output_buffer,
            kmip_version=kmip_version
        )
        output_buffer.write(local_buffer.buffer)

    def __eq__(self, other):
        if isinstance(other, LocateResponsePayload):
            if self.located_items != other.located_items:
                return False
            elif self.unique_identifiers != other.unique_identifiers:
                return False
            else:
                return True
        else:
            return NotImplemented

    def __ne__(self, other):
        if isinstance(other, LocateResponsePayload):
            return not (self == other)
        else:
            return NotImplemented

    def __repr__(self):
        args = ", ".join([
            "located_items={}".format(self.located_items),
            "unique_identifiers={}".format(self.unique_identifiers)
        ])
        return "LocateResponsePayload({})".format(args)

    def __str__(self):
        value = ", ".join(
            [
                '"located_items": {}'.format(self.located_items),
                '"unique_identifiers": {}'.format(self.unique_identifiers),
            ]
        )
        return '{' + value + '}'



# Copyright (c) 2016 The Johns Hopkins University/Applied Physics Laboratory
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

try:
    import unittest.mock as mock
except Exception:
    import mock

import signal
import socket
import testtools

from kmip.core import exceptions
from kmip.services import auth
from kmip.services.server import server


class TestKmipServer(testtools.TestCase):
    """
    A test suite for the KmipServer.
    """

    def setUp(self):
        super(TestKmipServer, self).setUp()

    def tearDown(self):
        super(TestKmipServer, self).tearDown()

    @mock.patch('kmip.services.server.server.KmipServer._setup_logging')
    @mock.patch('kmip.services.server.server.KmipServer._setup_configuration')
    def test_init(self, config_mock, logging_mock):
        """
        Test that a KmipServer can be instantiated without error.
        """
        s = server.KmipServer()
        self.assertTrue(config_mock.called)
        self.assertTrue(logging_mock.called)

        self.assertIsInstance(s.auth_suite, auth.BasicAuthenticationSuite)
        self.assertEqual(1, s._session_id)
        self.assertFalse(s._is_serving)

    @mock.patch('logging.getLogger', side_effect=mock.MagicMock())
    @mock.patch('logging.handlers.RotatingFileHandler')
    @mock.patch('kmip.services.server.server.KmipServer._setup_configuration')
    @mock.patch('os.path.exists')
    @mock.patch('os.path.isdir')
    @mock.patch('os.makedirs')
    def test_setup_logging(
            self,
            makedirs_mock,
            isdir_mock,
            path_mock,
            config_mock,
            handler_mock,
            logging_mock):
        """
        Verify that the server logger is setup correctly.
        """
        path_mock.return_value = False
        isdir_mock.return_value = False
        open_mock = mock.mock_open()

        # Dynamically mock out the built-in open function. Approach changes
        # across Python versions.
        try:
            # For Python3+
            import builtins  # NOQA
            module = 'builtins'
        except ImportError:
            # For Python2+
            module = '__builtin__'

        with mock.patch('{0}.open'.format(module), open_mock):
            s = server.KmipServer(log_path='/test/path/server.log')

        path_mock.assert_called_once_with('/test/path/server.log')
        isdir_mock.assert_called_once_with('/test/path')
        makedirs_mock.assert_called_once_with('/test/path')
        open_mock.assert_called_once_with('/test/path/server.log', 'w')

        self.assertTrue(s._logger.addHandler.called)
        s._logger.setLevel.assert_any_call(logging.DEBUG)

    @mock.patch('kmip.services.server.engine.KmipEngine')
    @mock.patch('kmip.services.auth.TLS12AuthenticationSuite')
    @mock.patch('kmip.services.server.server.KmipServer._setup_logging')
    def test_setup_configuration(self, logging_mock, auth_mock, engine_mock):
        """
        Test that the server setup configuration works without error.
        """
        s = server.KmipServer(
            config_path=None,
            policy_path=None
        )
        s.config = mock.MagicMock()

        # Test the right calls are made when reinvoking config setup
        s._setup_configuration(
            '/etc/pykmip/server.conf',
            '127.0.0.1',
            5696,
            '/etc/pykmip/certs/server.crt',
            '/etc/pykmip/certs/server.key',
            '/etc/pykmip/certs/ca.crt',
            'Basic',
            '/etc/pykmip/policies',
            False,
            'TLS_RSA_WITH_AES_128_CBC_SHA,TLS_RSA_WITH_AES_256_CBC_SHA',
            'DEBUG',
            '/var/pykmip/pykmip.db'
        )

        s.config.load_settings.assert_called_with('/etc/pykmip/server.conf')
        s.config.set_setting.assert_any_call('hostname', '127.0.0.1')
        s.config.set_setting.assert_any_call('port', 5696)
        s.config.set_setting.assert_any_call(
            'certificate_path',
            '/etc/pykmip/certs/server.crt'
        )
        s.config.set_setting.assert_any_call(
            'key_path',
            '/etc/pykmip/certs/server.key'
        )
        s.config.set_setting.assert_any_call(
            'ca_path',
            '/etc/pykmip/certs/ca.crt'
        )
        s.config.set_setting.assert_any_call('auth_suite', 'Basic')
        s.config.set_setting.assert_any_call(
            'policy_path',
            '/etc/pykmip/policies'
        )
        s.config.set_setting.assert_any_call(
            'enable_tls_client_auth',
            False
        )
        s.config.set_setting.assert_any_call(
            'tls_cipher_suites',
            [
                'TLS_RSA_WITH_AES_128_CBC_SHA',
                'TLS_RSA_WITH_AES_256_CBC_SHA'
            ]
        )
        s.config.set_setting.assert_any_call('logging_level', 'DEBUG')
        s.config.set_setting.assert_any_call(
            'database_path',
            '/var/pykmip/pykmip.db'
        )

        # Test that an attempt is made to instantiate the TLS 1.2 auth suite
        s = server.KmipServer(
            auth_suite='TLS1.2',
            config_path=None,
            policy_path=None
        )
        self.assertEqual('TLS1.2', s.config.settings.get('auth_suite'))
        self.assertIsNotNone(s.auth_suite)

    @mock.patch('multiprocessing.Manager')
    @mock.patch('kmip.services.server.monitor.PolicyDirectoryMonitor')
    @mock.patch('kmip.services.server.engine.KmipEngine')
    @mock.patch('kmip.services.server.server.KmipServer._setup_logging')
    def test_start(self,
                   logging_mock,
                   engine_mock,
                   monitor_mock,
                   manager_mock):
        """
        Test that starting the KmipServer either runs as expected or generates
        the expected error.
        """
        monitor_instance_mock = mock.MagicMock()
        monitor_mock.return_value = monitor_instance_mock

        dict_mock = mock.MagicMock()
        manager_instance_mock = mock.MagicMock()
        manager_instance_mock.dict.return_value = dict_mock
        manager_mock.return_value = manager_instance_mock

        a_mock = mock.MagicMock()
        b_mock = mock.MagicMock()

        s = server.KmipServer(
            hostname='127.0.0.1',
            port=5696,
            auth_suite='Basic',
            config_path=None,
            policy_path=None,
            tls_cipher_suites='TLS_RSA_WITH_AES_128_CBC_SHA'
        )
        s._logger = mock.MagicMock()

        self.assertFalse(s._is_serving)

        # Test that in ideal cases no errors are generated and the right
        # log messages are.
        with mock.patch('socket.socket') as socket_mock:
            with mock.patch('ssl.wrap_socket') as ssl_mock:
                socket_mock.return_value = a_mock
                ssl_mock.return_value = b_mock

                manager_mock.assert_not_called()
                monitor_mock.assert_not_called()

                s.start()

                manager_mock.assert_called_once_with()
                monitor_mock.assert_called_once_with(
                    None,
                    dict_mock,
                    False
                )
                self.assertIsNotNone(s._engine)
                s._logger.info.assert_any_call(
                    "Starting server socket handler."
                )
                s._logger.debug.assert_any_call("Configured cipher suites: 1")
                s._logger.debug.assert_any_call("TLS_RSA_WITH_AES_128_CBC_SHA")
                s._logger.debug.assert_any_call(
                    "Authentication suite ciphers to use: 1"
                )
                s._logger.debug.assert_any_call("AES128-SHA")

                socket_mock.assert_called_once_with(
                    socket.AF_INET,
                    socket.SOCK_STREAM
                )
                a_mock.setsockopt.assert_called_once_with(
                    socket.SOL_SOCKET,
                    socket.SO_REUSEADDR,
                    1
                )
                self.assertTrue(ssl_mock.called)
                b_mock.bind.assert_called_once_with(('127.0.0.1', 5696))
                s._logger.info.assert_called_with(
                    "Server successfully bound socket handler to "
                    "127.0.0.1:5696"
                )

        monitor_instance_mock.stop.assert_not_called()
        handler = signal.getsignal(signal.SIGINT)
        handler(None, None)
        monitor_instance_mock.stop.assert_called_once_with()
        monitor_instance_mock.stop.reset_mock()
        monitor_instance_mock.stop.assert_not_called()
        handler = signal.getsignal(signal.SIGTERM)
        handler(None, None)
        monitor_instance_mock.stop.assert_called_once_with()

        self.assertTrue(s._is_serving)

        manager_mock.reset_mock()
        monitor_mock.reset_mock()
        a_mock.reset_mock()
        b_mock.reset_mock()

        # Test that a NetworkingError is generated if the socket bind fails.
        with mock.patch('socket.socket') as socket_mock:
            with mock.patch('ssl.wrap_socket') as ssl_mock:
                socket_mock.return_value = a_mock
                ssl_mock.return_value = b_mock

                test_exception = Exception()
                b_mock.bind.side_effect = test_exception

                manager_mock.assert_not_called()
                monitor_mock.assert_not_called()

                regex = (
                    "Server failed to bind socket handler to 127.0.0.1:5696"
                )
                self.assertRaisesRegex(
                    exceptions.NetworkingError,
                    regex,
                    s.start
                )

                manager_mock.assert_called_once_with()
                monitor_mock.assert_called_once_with(
                    None,
                    dict_mock,
                    False
                )
                s._logger.info.assert_any_call(
                    "Starting server socket handler."
                )
                s._logger.exception.assert_called_once_with(test_exception)

    @mock.patch('kmip.services.server.engine.KmipEngine')
    @mock.patch('kmip.services.server.server.KmipServer._setup_logging')
    def test_stop(self, logging_mock, engine_mock):
        """
        Test that the right calls and log messages are triggered while
        cleaning up the server and any remaining sessions.
        """
        s = server.KmipServer(
            hostname='127.0.0.1',
            port=5696,
            config_path=None,
            policy_path=None
        )
        s._logger = mock.MagicMock()
        s._socket = mock.MagicMock()

        # Test the expected behavior for a normal server stop sequence
        thread_mock = mock.MagicMock()
        thread_mock.join = mock.MagicMock()
        thread_mock.is_alive = mock.MagicMock(return_value=False)
        thread_mock.name = 'TestThread'

        with mock.patch('threading.enumerate') as threading_mock:
            threading_mock.return_value = [thread_mock]

            s.stop()
            s._logger.info.assert_any_call(
                "Cleaning up remaining connection threads."
            )
            self.assertTrue(threading_mock.called)
            thread_mock.join.assert_called_once_with(10.0)
            s._logger.info.assert_any_call(
                "Cleanup succeeded for thread: TestThread"
            )
            s._logger.info.assert_any_call(
                "Shutting down server socket handler."
            )
            s._socket.shutdown.assert_called_once_with(socket.SHUT_RDWR)
            s._socket.close.assert_called_once_with()

        # Test the expected behavior when stopping multiple server session
        # threads goes wrong
        thread_mock.reset_mock()
        test_exception = Exception()
        thread_mock.join = mock.MagicMock(side_effect=test_exception)

        s._logger.reset_mock()
        s._socket.reset_mock()

        with mock.patch('threading.enumerate') as threading_mock:
            threading_mock.return_value = [thread_mock]

            s.stop()
            s._logger.info.assert_any_call(
                "Cleaning up remaining connection threads."
            )
            self.assertTrue(threading_mock.called)
            thread_mock.join.assert_called_once_with(10.0)
            s._logger.info.assert_any_call(
                "Error occurred while attempting to cleanup thread: TestThread"
            )
            s._logger.exception.assert_called_once_with(test_exception)
            s._logger.info.assert_any_call(
                "Shutting down server socket handler."
            )
            s._socket.shutdown.assert_called_once_with(socket.SHUT_RDWR)
            s._socket.close.assert_called_once_with()

        thread_mock.reset_mock()
        test_exception = Exception()
        thread_mock.join = mock.MagicMock()
        thread_mock.is_alive = mock.MagicMock(return_value=True)

        s._logger.reset_mock()
        s._socket.reset_mock()

        with mock.patch('threading.enumerate') as threading_mock:
            threading_mock.return_value = [thread_mock]

            s.stop()
            s._logger.info.assert_any_call(
                "Cleaning up remaining connection threads."
            )
            self.assertTrue(threading_mock.called)
            thread_mock.join.assert_called_once_with(10.0)
            s._logger.warning.assert_any_call(
                "Cleanup failed for thread: TestThread. Thread is still alive"
            )
            s._logger.info.assert_any_call(
                "Shutting down server socket handler."
            )
            s._socket.shutdown.assert_called_once_with(socket.SHUT_RDWR)
            s._socket.close.assert_called_once_with()

        # Test that the right errors and log messages are generated when
        # stopping the server goes wrong
        s._logger.reset_mock()
        s._socket.reset_mock()

        test_exception = Exception()
        s._socket.close = mock.MagicMock(side_effect=test_exception)

        regex = "Server failed to shutdown socket handler."
        self.assertRaisesRegex(
            exceptions.NetworkingError,
            regex,
            s.stop
        )
        s._logger.info.assert_any_call(
            "Cleaning up remaining connection threads."
        )
        s._logger.info.assert_any_call(
            "Shutting down server socket handler."
        )
        s._socket.shutdown.assert_called_once_with(socket.SHUT_RDWR)
        s._socket.close.assert_called_once_with()
        s._logger.exception(test_exception)

    @mock.patch('kmip.services.server.engine.KmipEngine')
    @mock.patch('kmip.services.server.server.KmipServer._setup_logging')
    def test_stop_with_monitor_shutdown_error(self, logging_mock, engine_mock):
        """
        Test that the right calls and log messages are triggered when stopping
        the server results in an error while shutting down the policy monitor.
        """
        s = server.KmipServer(
            hostname='127.0.0.1',
            port=5696,
            config_path=None,
            policy_path=None
        )
        s._logger = mock.MagicMock()
        s._socket = mock.MagicMock()
        s.policy_monitor = mock.MagicMock()
        test_exception = Exception()
        s.policy_monitor.join.side_effect = test_exception

        # Test the expected behavior for a normal server stop sequence
        thread_mock = mock.MagicMock()
        thread_mock.join = mock.MagicMock()
        thread_mock.is_alive = mock.MagicMock(return_value=False)
        thread_mock.name = 'TestThread'

        regex = "Server failed to clean up the policy monitor."
        self.assertRaisesRegex(
            exceptions.ShutdownError,
            regex,
            s.stop
        )
        s._logger.info.assert_any_call(
            "Cleaning up remaining connection threads."
        )
        s._logger.info.assert_any_call(
            "Shutting down server socket handler."
        )
        s._socket.shutdown.assert_called_once_with(socket.SHUT_RDWR)
        s._socket.close.assert_called_once_with()

        s.policy_monitor.stop.assert_called_once_with()
        s.policy_monitor.join.assert_called_once_with()
        s._logger.exception(test_exception)

    @mock.patch('kmip.services.server.engine.KmipEngine')
    @mock.patch('kmip.services.server.server.KmipServer._setup_logging')
    def test_serve(self, logging_mock, engine_mock):
        """
        Test that the right calls and log messages are triggered while
        serving connections.
        """
        s = server.KmipServer(
            hostname='127.0.0.1',
            port=5696,
            config_path=None,
            policy_path=None
        )
        s._is_serving = True
        s._logger = mock.MagicMock()
        s._socket = mock.MagicMock()
        s._setup_connection_handler = mock.MagicMock()

        expected_error = KeyboardInterrupt

        # Test the expected behavior for a normal server/interrupt sequence
        s._socket.accept = mock.MagicMock(
            side_effect=[
                ('connection', 'address'),
                socket.timeout,
                expected_error
            ]
        )

        s.serve()
        s._socket.listen.assert_called_once_with(5)
        s._socket.accept.assert_any_call()
        s._setup_connection_handler.assert_called_once_with(
            'connection',
            'address'
        )
        s._logger.warning.assert_called_with(
            "Interrupting connection service."
        )
        s._logger.info.assert_called_with("Stopping connection service.")

        # Test the behavior for an unexpected socket error.
        unexpected_error = socket.error()
        s._is_serving = True
        s._logger.reset_mock()
        s._socket.accept = mock.MagicMock(
            side_effect=[unexpected_error, expected_error]
        )

        s.serve()
        s._socket.accept.assert_any_call()
        s._logger.warning.assert_any_call(
            "Error detected while establishing new connection."
        )
        s._logger.exception.assert_called_with(unexpected_error)
        s._logger.info.assert_called_with("Stopping connection service.")

        # Test the behavior for an unexpected error.
        unexpected_error = Exception()
        s._is_serving = True
        s._logger.reset_mock()
        s._socket.accept = mock.MagicMock(
            side_effect=[unexpected_error, expected_error]
        )

        s.serve()
        s._socket.accept.assert_any_call()
        s._logger.warning.assert_any_call(
            "Error detected while establishing new connection."
        )
        s._logger.exception.assert_called_with(unexpected_error)
        s._logger.info.assert_called_with("Stopping connection service.")

        # Test the signal handler for each expected signal
        s._is_serving = True
        handler = signal.getsignal(signal.SIGINT)
        args = (signal.SIGINT, None)
        self.assertRaisesRegex(
            KeyboardInterrupt,
            "SIGINT received",
            handler,
            *args
        )
        self.assertFalse(s._is_serving)

        s._is_serving = True
        handler = signal.getsignal(signal.SIGTERM)
        handler(None, None)
        self.assertFalse(s._is_serving)

    @mock.patch('kmip.services.server.engine.KmipEngine')
    @mock.patch('kmip.services.server.server.KmipServer._setup_logging')
    def test_setup_connection_handler(self, logging_mock, engine_mock):
        """
        Test that a KmipSession can be successfully created and spun off from
        the KmipServer.
        """
        s = server.KmipServer(
            hostname='127.0.0.1',
            port=5696,
            config_path=None,
            policy_path=None
        )
        s._logger = mock.MagicMock()
        s._engine = engine_mock

        # Test that the right calls and log messages are made when
        # starting a new session.
        with mock.patch(
            'kmip.services.server.session.KmipSession.start'
        ) as session_mock:
            address = ('127.0.0.1', 5696)
            s._setup_connection_handler(None, address)

            s._logger.info.assert_any_call(
                "Receiving incoming connection from: 127.0.0.1:5696"
            )
            s._logger.info.assert_any_call(
                "Dedicating session 00000001 to 127.0.0.1:5696"
            )
            session_mock.assert_called_once_with()

        self.assertEqual(2, s._session_id)

        # Test that the right error messages are logged when the session
        # fails to start.
        test_exception = Exception()
        with mock.patch(
            'kmip.services.server.session.KmipSession.start',
            side_effect=test_exception
        ) as session_mock:
            address = ('127.0.0.1', 5696)
            s._setup_connection_handler(None, address)

            s._logger.info.assert_any_call(
                "Receiving incoming connection from: 127.0.0.1:5696"
            )
            s._logger.info.assert_any_call(
                "Dedicating session 00000001 to 127.0.0.1:5696"
            )
            session_mock.assert_called_once_with()
            s._logger.warning.assert_called_once_with(
                "Failure occurred while starting session: 00000002"
            )
            s._logger.exception.assert_called_once_with(test_exception)

        self.assertEqual(3, s._session_id)

    @mock.patch('kmip.services.server.engine.KmipEngine')
    @mock.patch('kmip.services.server.server.KmipServer._setup_logging')
    def test_as_context_manager(self, logging_mock, engine_mock):
        """
        Test that the right methods are called when the KmipServer is used
        as a context manager.
        """
        s = server.KmipServer(
            hostname='127.0.0.1',
            port=5696,
            config_path=None,
            policy_path=None
        )
        s._logger = mock.MagicMock()
        s.start = mock.MagicMock()
        s.stop = mock.MagicMock()

        with s:
            pass

        self.assertTrue(s.start.called)
        self.assertTrue(s.stop.called)
